import logging

__all__ = ["algorithms", "kernel", "systems", "utils"]

logger = logging.getLogger(__name__)
"""Default logger for gym_socks.

Used mainly for debugging. Can be diabled entirely by setting the log level of the
logger to "notset".

"""

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(name)s - %(message)s")



__all__ = [
    # Systems
    "cwh",
    "integrator",
    "nonholonomic",
    "point_mass",
    "QUAD20",
    "tora",
    # Classes
    "policy",
    "sample",
]

from gym_socks.envs.cwh import CWH4DEnv
from gym_socks.envs.cwh import CWH6DEnv

from gym_socks.envs.integrator import NDIntegratorEnv

from gym_socks.envs.nonholonomic import NonholonomicVehicleEnv

from gym_socks.envs.planar_quad import PlanarQuadrotorEnv

from gym_socks.envs.point_mass import NDPointMassEnv

from gym_socks.envs.QUAD20 import QuadrotorEnv

from gym_socks.envs.tora import TORAEnv

# ---- Register gym envs ----

from gym.envs.registration import register

register(
    id="CWH4DEnv-v0",
    entry_point="gym_socks.envs:CWH4DEnv",
    order_enforce=False,
)

register(
    id="CWH6DEnv-v0",
    entry_point="gym_socks.envs:CWH6DEnv",
    order_enforce=False,
)

register(
    id="2DIntegratorEnv-v0",
    entry_point=NDIntegratorEnv,
    kwargs={"dim": 2},
    order_enforce=False,
)

register(
    id="NonholonomicVehicleEnv-v0",
    entry_point="gym_socks.envs:NonholonomicVehicleEnv",
    order_enforce=False,
)

register(
    id="PlanarQuadrotorEnv-v0",
    entry_point="gym_socks.envs:PlanarQuadrotorEnv",
    order_enforce=False,
)

register(
    id="2DPointMassEnv-v0",
    entry_point=NDPointMassEnv,
    kwargs={"dim": 2},
    order_enforce=False,
)

register(
    id="QuadrotorEnv-v0",
    entry_point="gym_socks.envs:QuadrotorEnv",
    order_enforce=False,
)

register(
    id="TORAEnv-v0",
    entry_point="gym_socks.envs:TORAEnv",
    order_enforce=False,
)



import unittest
from unittest import mock
from unittest.mock import patch

import gym

import numpy as np

from examples.ingredients.common import parse_array
from examples.ingredients.common import box_factory
from examples.ingredients.common import grid_ranges
from examples.ingredients.common import grid_sample_size


class TestParseConfigFunctions(unittest.TestCase):
    def test_parse_array_scalar(cls):
        """Test the parse_array function for scalar inputs."""

        groundTruth = [5, 5]

        result = parse_array(value=5, shape=(2,), dtype=np.float32)
        cls.assertTrue(np.array_equal(result, groundTruth))
        cls.assertTrue(np.array_equal(result, np.array(groundTruth)))
        cls.assertTrue(np.array_equal(result, np.array(groundTruth, dtype=np.float32)))

        groundTruth = [[5, 5], [5, 5]]

        result = parse_array(value=5, shape=(2, 2), dtype=np.float32)
        cls.assertTrue(np.array_equal(result, groundTruth))
        cls.assertTrue(np.array_equal(result, np.array(groundTruth)))
        cls.assertTrue(np.array_equal(result, np.array(groundTruth, dtype=np.float32)))

    def test_parse_array_list(cls):
        """Test the parse_array function for list inputs."""

        result = parse_array(value=[1, 2], shape=(2,), dtype=np.float32)
        cls.assertTrue(np.array_equal(result, [1, 2]))
        cls.assertTrue(np.array_equal(result, np.array([1, 2])))
        cls.assertTrue(np.array_equal(result, np.array([1, 2], dtype=np.float32)))

        result = parse_array(value=[[1, 2], [3, 4]], shape=(2, 2), dtype=np.float32)
        cls.assertTrue(np.array_equal(result, [[1, 2], [3, 4]]))
        cls.assertTrue(np.array_equal(result, np.array([[1, 2], [3, 4]])))
        cls.assertTrue(
            np.array_equal(result, np.array([[1, 2], [3, 4]], dtype=np.float32))
        )


class TestBoxFactory(unittest.TestCase):
    def test_box_factory(cls):
        """Test box factory produces correct spaces."""

        result = box_factory(
            lower_bound=-1, upper_bound=1, shape=(2,), dtype=np.float32
        )

        cls.assertTrue(np.array_equal(result.low, [-1, -1]))
        cls.assertTrue(np.array_equal(result.high, [1, 1]))

        cls.assertEqual(result.shape, (2,))

        cls.assertIsInstance(result, gym.spaces.Box)

        result = box_factory(
            lower_bound=[-1, -2], upper_bound=[1, 2], shape=(2,), dtype=np.float32
        )

        cls.assertTrue(np.array_equal(result.low, [-1, -2]))
        cls.assertTrue(np.array_equal(result.high, [1, 2]))

        cls.assertEqual(result.shape, (2,))

        cls.assertIsInstance(result, gym.spaces.Box)


class TestGridRanges(unittest.TestCase):
    def test_grid_ranges(cls):
        """Test that grid_ranges produces the correct ranges."""

        space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        result = grid_ranges(space=space, grid_resolution=5)

        groundTruth = [np.linspace(-1, 1, 5), np.linspace(-1, 1, 5)]

        cls.assertTrue(np.array_equal(result, groundTruth))

        groundTruth = np.array([np.linspace(-1, 1, 5), np.linspace(-1, 1, 5)])

        cls.assertTrue(np.array_equal(result, groundTruth))


class TestGridSampleSize(unittest.TestCase):
    def test_grid_sample_size(cls):
        """Test that grid_sample_size computes the correct sample size."""

        space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        result = grid_sample_size(space=space, grid_resolution=5)

        groundTruth = 25

        cls.assertEqual(result, groundTruth)



"""Backward in time stochastic optimal control.

References:
    .. [1] `Stochastic Optimal Control via
            Hilbert Space Embeddings of Distributions, 2021
            Adam J. Thorpe, Meeko M. K. Oishi
            IEEE Conference on Decision and Control,
            <https://arxiv.org/abs/2103.12759>`_

"""

from functools import partial

import gym_socks

from gym_socks.envs.policy import BasePolicy
from gym_socks.algorithms.control.control_common import compute_solution
from gym_socks.kernel.metrics import rbf_kernel, regularized_inverse
from gym_socks.envs.sample import transpose_sample
from gym_socks.utils import generate_batches
from gym_socks.utils.logging import ms_tqdm, _progress_fmt

import numpy as np


def _compute_backward_recursion(
    Y,
    W,
    CXY,
    CUA,
    num_steps=None,
    cost_fn=None,
    constraint_fn=None,
    value_functions=None,
    out=None,
    verbose=False,
):
    # Compute beta.
    gym_socks.logger.debug("Computing beta.")
    beta = np.einsum("ij,ik->ijk", CXY, CUA)

    pbar = ms_tqdm(
        total=num_steps,
        bar_format=_progress_fmt,
        disable=False if verbose is True else True,
    )

    # set up empty array to hold value functions
    if out is None:
        out = np.zeros((num_steps, len(Y)), dtype=np.float32)
    out[num_steps - 1, :] = np.array(cost_fn(time=num_steps - 1), dtype=np.float32)
    pbar.update()

    # run backwards in time and compute the safety probabilities
    for t in range(num_steps - 2, -1, -1):

        Z = np.matmul(value_functions[t + 1, :], W)
        C = np.einsum("i,ijk->jk", Z, beta)

        V = np.zeros((len(Y),))

        if constraint_fn is not None:
            R = np.matmul(constraint_fn(time=t), W)
            D = np.einsum("i,ijk->jk", R, beta)

            for i in range(len(Y)):
                CA = C[i][D[i] <= 0]
                if CA.size == 0:
                    V[i] = np.min(C[i])
                else:
                    V[i] = np.min(CA)

        else:
            V = np.min(C, axis=1)

        # for i in range(len(Y)):
        #     sol = compute_solution(C[i], D[i])
        #     # idx = np.argmax(sol)
        #     # V =
        #     V[i] = sol

        out[t, :] = cost_fn(time=t) + V

        pbar.update()

    pbar.close()

    return out


def _compute_backward_recursion_batch(
    Y,
    W,
    CXY,
    CUA,
    num_steps=None,
    cost_fn=None,
    constraint_fn=None,
    value_functions=None,
    out=None,
    batch_size=None,
    verbose=False,
):
    pbar = ms_tqdm(
        total=num_steps,
        bar_format=_progress_fmt,
        disable=False if verbose is True else True,
    )

    # set up empty array to hold value functions
    if out is None:
        out = np.zeros((num_steps + 1, len(Y)), dtype=np.float32)
    out[num_steps - 1, :] = np.array(cost_fn(time=num_steps - 1), dtype=np.float32)
    pbar.update()

    # run backwards in time and compute the safety probabilities
    for t in range(num_steps - 2, -1, -1):

        Z = np.matmul(value_functions[t + 1, :], W)
        # C = np.empty((1, len(Y)))
        C = np.empty_like(CUA)

        # if constraint_fn is not None:
        # D = np.empty((1, len(Y)))
        D = np.empty_like(CUA)

        for batch in generate_batches(num_elements=len(Y), batch_size=batch_size):

            beta = np.einsum("ij,ik->ijk", CXY[:, batch], CUA)
            C[batch, :] = np.einsum("i,ijk->jk", Z, beta)

            if constraint_fn is not None:
                R = np.matmul(constraint_fn(time=t), W)
                D[batch, :] = np.einsum("i,ijk->jk", R, beta)

        # V = np.zeros((len(Y),))
        # for i in range(len(Y)):
        #     # sol = compute_solution(C[i], D[i])
        #     # idx = np.argmax(sol)
        #     # # print(f"V shape {V.shape}")
        #     # # print(f"V shape {V[i].shape}")
        #     # # print(f"C chape {C.shape}")
        #     # V[i] = C[:, idx]

        #     CA = C[i][D[i] <= 0]
        #     if CA.size == 0:
        #         V[i] = np.min(C[i])
        #     else:
        #         V[i] = np.min(CA)

        #     #     out[t, :] = cost_fn(time=t) + V

        #     # else:

        #     #     Z = np.matmul(value_functions[t + 1, :], W)

        #     #     C = np.zeros_like(CUA)

        #     #     for batch in generate_batches(num_elements=len(Y), batch_size=batch_size):

        #     #         beta = np.einsum("ij,ik->ijk", CXY[:, batch], CUA)
        #     #         C[batch, :] = np.einsum("i,ijk->jk", Z, beta)

        #     #     V = np.min(C, axis=1)

        if constraint_fn is not None:

            V = np.zeros((len(Y),))

            for i in range(len(Y)):

                CA = C[i][D[i] <= 0]
                if CA.size == 0:
                    V[i] = np.min(C[i])
                else:
                    V[i] = np.min(CA)
        else:
            V = np.min(C, axis=1)

        out[t, :] = cost_fn(time=t) + V

        pbar.update()

    pbar.close()

    return out


def kernel_control_bwd(
    S: np.ndarray,
    A: np.ndarray,
    num_steps: int = None,
    cost_fn=None,
    constraint_fn=None,
    heuristic: bool = False,
    regularization_param: float = None,
    kernel_fn=None,
    batch_size: int = None,
    verbose: bool = True,
):
    """Stochastic optimal control policy backward in time.

    Computes the optimal policy using dynamic programming. The solution computes an approximation of the value functions starting at the terminal time and working backwards. Then, when the policy is evaluated, it moves forward in time, optimizing over the value functions and choosing the action which has the highest "value".

    Args:
        S: Sample taken iid from the system evolution.
        A: Collection of admissible control actions.
        cost_fn: The cost function. Should return a real value.
        constraint_fn: The constraint function. Should return a real value.
        heuristic: Whether to use the heuristic solution instead of solving the LP.
        regularization_param: Regularization prameter for the regularized least-squares
            problem used to construct the approximation.
        kernel_fn: The kernel function used by the algorithm.
        verbose: Whether the algorithm should print verbose output.

    """

    alg = KernelControlBwd(
        num_steps=num_steps,
        cost_fn=cost_fn,
        constraint_fn=constraint_fn,
        heuristic=heuristic,
        regularization_param=regularization_param,
        kernel_fn=kernel_fn,
        batch_size=batch_size,
        verbose=verbose,
    )
    alg.train(S, A)

    return alg


class KernelControlBwd(BasePolicy):
    """Stochastic optimal control policy backward in time.

    Computes the optimal policy using dynamic programming. The solution computes an approximation of the value functions starting at the terminal time and working backwards. Then, when the policy is evaluated, it moves forward in time, optimizing over the value functions and choosing the action which has the highest "value".

    Args:
        S: Sample taken iid from the system evolution.
        A: Collection of admissible control actions.
        cost_fn: The cost function. Should return a real value.
        constraint_fn: The constraint function. Should return a real value.
        heuristic: Whether to use the heuristic solution instead of solving the LP.
        regularization_param: Regularization prameter for the regularized least-squares
            problem used to construct the approximation.
        kernel_fn: The kernel function used by the algorithm.
        verbose: Whether the algorithm should print verbose output.

    """

    def __init__(
        self,
        num_steps: int = None,
        cost_fn=None,
        constraint_fn=None,
        heuristic: bool = False,
        regularization_param: float = None,
        kernel_fn=None,
        batch_size: int = None,
        verbose: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.num_steps = num_steps

        self.cost_fn = cost_fn
        self.constraint_fn = constraint_fn

        self.heuristic = heuristic

        self.kernel_fn = kernel_fn
        self.regularization_param = regularization_param

        self.batch_size = batch_size

        self.verbose = verbose

    def _validate_params(self, S=None, A=None):

        if self.kernel_fn is None:
            self.kernel_fn = partial(rbf_kernel, sigma=0.1)

        if self.regularization_param is None:
            self.regularization_param = 1

        if S is None:
            raise ValueError("Must supply a sample.")

        if A is None:
            raise ValueError("Must supply a sample.")

        if self.cost_fn is None:
            raise ValueError("Must supply a cost function.")

        if self.constraint_fn is None:
            # raise ValueError("Must supply a constraint function.")
            self._constrained = False
        else:
            self._constrained = True

        if self.batch_size is not None:
            self._compute_backward_recursion_caller = partial(
                _compute_backward_recursion_batch, batch_size=self.batch_size
            )

        else:
            self._compute_backward_recursion_caller = _compute_backward_recursion

    def _validate_data(self, S):

        if S is None:
            raise ValueError("Must supply a sample.")

    def train(self, S: np.ndarray, A: np.ndarray):
        """Train the algorithm.

        Args:
            S: Sample taken iid from the system evolution.
            A: Collection of admissible control actions.

        Returns:
            self: An instance of the KernelControlFwd algorithm class.

        """

        self._validate_data(S)
        self._validate_data(A)
        self._validate_params(S, A)

        X, U, Y = transpose_sample(S)
        X = np.array(X)
        U = np.array(U)
        Y = np.array(Y)

        self.X = X

        A = np.array(A)

        self.A = A

        gym_socks.logger.debug("Computing matrix inverse.")
        self.W = regularized_inverse(
            X,
            U=U,
            kernel_fn=self.kernel_fn,
            regularization_param=self.regularization_param,
        )

        gym_socks.logger.debug("Computing covariance matrices.")
        self.CUA = self.kernel_fn(U, A)
        CXY = self.kernel_fn(X, Y)

        gym_socks.logger.debug("Computing cost function.")
        self.cost_fn = partial(self.cost_fn, state=Y)

        gym_socks.logger.debug("Computing constraint function.")
        if self.constraint_fn is not None:
            self.constraint_fn = partial(self.constraint_fn, state=Y)

        gym_socks.logger.debug("Computing value functions.")
        value_functions = np.empty((self.num_steps, len(Y)))
        self.value_functions = self._compute_backward_recursion_caller(
            Y,
            self.W,
            CXY,
            self.CUA,
            num_steps=self.num_steps,
            cost_fn=self.cost_fn,
            constraint_fn=self.constraint_fn,
            value_functions=value_functions,
            out=value_functions,
            verbose=True,
        )

    def __call__(self, time=0, state=None, *args, **kwargs):

        if state is None:
            print("Must supply a state to the policy.")
            return None

        T = np.array(state)

        # Compute covariance matrix.
        CXT = self.kernel_fn(self.X, T)

        # Compute beta.
        betaXT = self.W @ (CXT * self.CUA)

        # Compute cost vector.
        C = np.matmul(np.array(self.value_functions[time], dtype=np.float32), betaXT)

        # Compute constraint vector.
        D = None
        if self.constraint_fn is not None:
            D = np.matmul(
                np.array(self.constraint_fn(time=time), dtype=np.float32), betaXT
            )

        # Compute the solution to the LP.
        gym_socks.logger.debug("Computing solution to the LP.")
        sol = compute_solution(C, D, heuristic=self.heuristic)
        idx = np.argmax(sol)
        return np.array(self.A[idx], dtype=np.float32)



"""Simulation ingredient.

Used for experiments where the system needs to be simulated, such as after computing a
policy via one of the included control algorithms.

"""

from sacred import Ingredient

from gym_socks.envs.dynamical_system import DynamicalSystem
from gym_socks.envs.policy import BasePolicy

import numpy as np

simulation_ingredient = Ingredient("simulation")


@simulation_ingredient.config
def config():
    """Simulation configuration variables."""

    initial_condition = None  # Initial condition of the system.

    filename = "results/data.npy"


@simulation_ingredient.capture
def simulate_system(
    env: DynamicalSystem, policy: BasePolicy, initial_condition: list, _log
) -> list:
    """Simulate the system from a given initial condition."""

    _log.debug("Simulating the system.")

    # Set the initial condition.
    env.reset()

    if initial_condition is None:
        initial_condition = env.state_space.sample()

    env.state = initial_condition

    trajectory = [env.state]

    # Simulate the env using the computed policy.
    for t in range(env.num_time_steps):

        action = np.array(policy(time=t, state=[env.state]), dtype=np.float32)
        obs, cost, done, _ = env.step(action)
        next_state = env.state

        trajectory.append(list(next_state))

        if done:
            break

    return trajectory



"""Planar quadrotor system."""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np

from scipy.constants import g
from scipy.integrate import solve_ivp


class PlanarQuadrotorEnv(DynamicalSystem):
    """Planar quadrotor system.

    A planar quadrotor is quadrotor restricted to two dimensions. Similar to the OpenAI gym lunar lander benchmark, the planar quadrotor is a bar with two independent rotors at either end. Inputs are the trust of the rotors, and apply a torque to the bar. The system is also subject to gravitational forces.

    """

    def __init__(self, *args, **kwargs):
        """Initialize the system."""
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(2,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        self._gravitational_acceleration = g  # [m/s^2]
        self._rotor_distance = 2  # [m]
        self._total_mass = 5  # [kg]
        self._inertia = 2

    @property
    def gravitational_acceleration(self):
        return self._gravitational_acceleration

    @gravitational_acceleration.setter
    def gravitational_acceleration(self, value):
        self._gravitational_acceleration = value

    @property
    def rotor_distance(self):
        return self._rotor_distance

    @rotor_distance.setter
    def rotor_distance(self, value):
        self._rotor_distance = value

    @property
    def total_mass(self):
        return self._total_mass

    @total_mass.setter
    def total_mass(self, value):
        self._total_mass = value

    @property
    def inertia(self):
        return self._inertia

    @inertia.setter
    def inertia(self, value):
        self._inertia = value

    def step(self, action, time=0):
        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        disturbance = self.generate_disturbance(time, self.state, action)

        # solve the initial value problem
        sol = solve_ivp(
            self.dynamics,
            [0, self.sampling_time],
            self.state,
            args=(
                action,
                disturbance,
            ),
        )
        *_, self.state = sol.y.T

        # correct the angle
        if np.abs(self.state[5]) >= 2 * np.pi:
            self.state[5] %= 2 * np.pi

        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        w = np.multiply([1e-3, 1e-5, 1e-3, 1e-5, 1e-3, 1e-5], w)
        return np.array(w)

    def dynamics(self, time, state, action, disturbance):
        """Dynamics for the system."""

        x1, x2, x3, x4, x5, x6 = state
        u1, u2 = action
        w1, w2, w3, w4, w5, w6 = disturbance

        M = u1 + u2

        dx1 = x2 + w1
        dx2 = -(M * np.sin(x5)) / self.total_mass + w2
        dx3 = x4 + w3
        dx4 = (M * np.cos(x5)) / self.total_mass - self.gravitational_acceleration + w4
        dx5 = x5 + w5
        dx6 = (self.rotor_distance * (u1 - u2)) / self.inertia + w6

        return np.array([dx1, dx2, dx3, dx4, dx5, dx6], dtype=np.float32)



from sacred import Ingredient

import gym

import numpy as np

from gym.envs.registration import make

from gym_socks.envs.dynamical_system import DynamicalSystem

from examples.ingredients.common import assert_config_has_key, box_factory

system_ingredient = Ingredient("system")


@system_ingredient.config
def _config():
    """System configuration variables."""

    system_id = "2DIntegratorEnv-v0"

    time_horizon = None
    sampling_time = None


@system_ingredient.capture
def make_system(
    system_id: str,
    time_horizon: float,
    sampling_time: float,
    _config,
    _log,
) -> DynamicalSystem:
    """Construct an instance of the system.

    Args:
        system_id: System identifier string.
        time_horizon: The time horizon of the system.
        sampling_time: The sampling time of the system.

    Returns:
        An instance of the dynamical system model.

    """

    _log.debug(f"Making system: {system_id}")
    env = make(system_id)

    if time_horizon is not None:
        env.time_horizon = time_horizon

    if sampling_time is not None:
        env.sampling_time = sampling_time

    if "action_space" in _config:
        _log.debug(f"Setting action space.")
        assert_config_has_key(_config["action_space"], "lower_bound")
        assert_config_has_key(_config["action_space"], "upper_bound")

        lower_bound = _config["action_space"]["lower_bound"]
        upper_bound = _config["action_space"]["upper_bound"]

        _log.debug(f"Action space lower bound: {lower_bound}")
        _log.debug(f"Action space upper bound: {upper_bound}")

        env.action_space = box_factory(
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            shape=env.action_space.shape,
            dtype=np.float32,
        )

    return env


@system_ingredient.capture
def set_system_seed(seed: int, env: DynamicalSystem, _log):
    """Set the random seed.

    Args:
        seed: The random seed for the experiment.
        env: The dynamical system model.

    """

    _log.debug(f"Setting seed to: {seed}")
    env.seed(seed=seed)
    env.observation_space.seed(seed=seed)
    env.state_space.seed(seed=seed)
    env.action_space.seed(seed=seed)


@system_ingredient.command(unobserved=True)
def print_info(system_id):
    """Print system info.

    Prints information of the system specified by `system_id` to the screen.

    Args:
        system_id: System identifier string.

    """

    env = make_system()

    from pprint import PrettyPrinter

    printer = PrettyPrinter(indent=4, width=88, compact=False, sort_dicts=False)

    print(env.__doc__)

    info = {
        "system_id": system_id,
        "state_space": {
            "shape": env.state_space.shape,
            "low": env.state_space.low,
            "high": env.state_space.high,
        },
        "action_space": {
            "shape": env.action_space.shape,
            "low": env.action_space.low,
            "high": env.action_space.high,
        },
        "observation_space": {
            "shape": env.observation_space.shape,
            "low": env.observation_space.low,
            "high": env.observation_space.high,
        },
    }

    printer.pprint(info)



import unittest
from unittest.mock import patch

import gym

import gym_socks.envs

import numpy as np


class TestToraSystem(unittest.TestCase):
    def test_set_damping_coefficient(cls):
        system = gym_socks.envs.TORAEnv()

        system.damping_coefficient = 0.5
        cls.assertEqual(system.damping_coefficient, 0.5)



from tqdm.auto import tqdm

_progress_fmt = "|{bar:30}| {elapsed_ms}"
"""The format of the tqdm progress bar.

The format of the progress bar used by the `ms_tqdm` custom progress bar class.
"""


class ms_tqdm(tqdm):
    """A custom tqdm progress bar implementation.

    This is a simple modification to the tqdm progress bar that shows millisecond
    computation times in the progress bar. Adds the `elapsed_ms` format specifier to
    the format string.

    """

    @property
    def format_dict(self):
        d = super(ms_tqdm, self).format_dict

        min, s = divmod(d["elapsed"], 60)
        h, m = divmod(min, 60)

        if h:
            elapsed_ms = "{0:d}:{1:02d}:{2:06.3f}".format(int(h), int(m), s)
        else:
            elapsed_ms = "{0:02d}:{1:06.3f}".format(int(m), s)

        d.update(elapsed_ms=elapsed_ms)
        return d



import unittest
from unittest.mock import patch

import gym

import gym_socks.envs

import numpy as np


class TestIntegratorSystem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.env = gym_socks.envs.NDIntegratorEnv(2)

    @patch.object(gym_socks.envs.NDIntegratorEnv, "generate_disturbance")
    def test_known_trajectory(cls, mock_generate_disturbance):
        """Test against specific known trajectory. Sanity check."""
        mock_generate_disturbance.return_value = np.zeros((2,))

        env = cls.env

        env.state = np.array([0.1, 0.1], dtype=np.float32)
        action = np.array([0], dtype=np.float32)

        trajectory = []
        trajectory.append(env.state)

        for i in range(env.num_time_steps):
            obs, cost, done, _ = env.step(action)
            trajectory.append(obs)

        trajectory = np.array(trajectory)

        groundTruth = np.array(
            [
                [0.100, 0.1],
                [0.125, 0.1],
                [0.150, 0.1],
                [0.175, 0.1],
                [0.200, 0.1],
                [0.225, 0.1],
                [0.250, 0.1],
                [0.275, 0.1],
                [0.300, 0.1],
                [0.325, 0.1],
                [0.350, 0.1],
                [0.375, 0.1],
                [0.400, 0.1],
                [0.425, 0.1],
                [0.450, 0.1],
                [0.475, 0.1],
                [0.500, 0.1],
            ]
        )

        cls.assertTrue(np.allclose(trajectory, groundTruth))

    @patch.object(gym_socks.envs.NDIntegratorEnv, "generate_disturbance")
    def test_euler_approximation(cls, mock_generate_disturbance):
        """Test against specific known trajectory (Euler). Sanity check."""
        mock_generate_disturbance.return_value = np.zeros((2,))

        env = cls.env
        env._euler = True

        env.state = np.array([0.1, 0.1], dtype=np.float32)
        action = np.array([0], dtype=np.float32)

        trajectory = []
        trajectory.append(env.state)

        for i in range(env.num_time_steps):
            obs, cost, done, _ = env.step(action)
            trajectory.append(obs)

        trajectory = np.array(trajectory)

        groundTruth = np.array(
            [
                [0.100, 0.1],
                [0.125, 0.1],
                [0.150, 0.1],
                [0.175, 0.1],
                [0.200, 0.1],
                [0.225, 0.1],
                [0.250, 0.1],
                [0.275, 0.1],
                [0.300, 0.1],
                [0.325, 0.1],
                [0.350, 0.1],
                [0.375, 0.1],
                [0.400, 0.1],
                [0.425, 0.1],
                [0.450, 0.1],
                [0.475, 0.1],
                [0.500, 0.1],
            ]
        )

        cls.assertTrue(np.allclose(trajectory, groundTruth))



"""Quadrotor system.

References:
    .. [1] `ARCH-COMP20 Category Report:
            Continuous and Hybrid Systems with Nonlinear Dynamics
            <https://easychair.org/publications/open/nrdD>`_

"""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np

from scipy.constants import g


class QuadrotorEnv(DynamicalSystem):
    """Quadrotor system.

    The quadrotor system is a high-dimensional (12D) system. The states are the
    position, velocity, and angles of the system, and the inputs are the torques on the
    angles.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(12,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(12,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(3,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        # system parameters
        self._gravitational_acceleration = g  # [m/s^2]
        self._radius_center_mass = 0.1  # [m]
        self._rotor_distance = 0.5  # [m]
        self._rotor_mass = 0.1  # [kg]
        self._center_mass = 1  # [kg]

        self._compute_total_mass()
        self._compute_inertia()

    @property
    def gravitational_acceleration(self):
        return self._gravitational_acceleration

    @gravitational_acceleration.setter
    def gravitational_acceleration(self, value):
        self._gravitational_acceleration = value

    @property
    def radius_center_mass(self):
        return self._radius_center_mass

    @radius_center_mass.setter
    def radius_center_mass(self, value):
        self._radius_center_mass = value
        self._compute_inertia()

    @property
    def rotor_distance(self):
        return self._rotor_distance

    @rotor_distance.setter
    def rotor_distance(self, value):
        self._rotor_distance = value
        self._compute_inertia()

    @property
    def rotor_mass(self):
        return self._rotor_mass

    @rotor_mass.setter
    def rotor_mass(self, value):
        self._rotor_mass = value
        self._compute_total_mass()
        self._compute_inertia()

    @property
    def center_mass(self):
        return self._center_mass

    @center_mass.setter
    def center_mass(self, value):
        self._center_mass = value
        self._compute_total_mass()
        self._compute_inertia()

    def _compute_total_mass(self):
        self.total_mass = self.center_mass + 4 * self.rotor_mass

    def _compute_inertia(self):
        self.Jx = (2 / 5) * self.center_mass * (self.radius_center_mass ** 2) + 2 * (
            self.rotor_distance ** 2
        ) * self.rotor_mass
        self.Jy = self.Jx
        self.Jz = (2 / 5) * self.center_mass * (self.radius_center_mass ** 2) + 4 * (
            self.rotor_distance ** 2
        ) * self.rotor_mass

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        return 1e-2 * np.array(w)

    def dynamics(self, time, state, action, disturbance):
        """Dynamics for the system."""
        x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12 = state
        u1, u2, u3 = action
        w1, w2, w3, w4, w5, w6, w7, w8, w9, w10, w11, w12 = disturbance

        F = self.total_mass * self.gravitational_acceleration - 10 * (x3 - u1) + 3 * x6
        tau_phi = -(x7 - u2) - x10
        tau_psi = -(x8 - u3) - x11
        tau_theta = 0

        dx1 = (
            np.cos(x8) * np.cos(x9) * x4
            + (np.sin(x7) * np.sin(x8) * np.cos(x9) - np.cos(x7) * np.sin(x9)) * x5
            + (np.cos(x7) * np.sin(x8) * np.cos(x9) + np.sin(x7) * np.sin(x9)) * x6
        ) + w1
        dx2 = (
            np.cos(x8) * np.sin(x9) * x4
            + (np.sin(x7) * np.sin(x8) * np.sin(x9) + np.cos(x7) * np.cos(x9)) * x5
            + (np.cos(x7) * np.sin(x8) * np.sin(x9) - np.sin(x7) * np.cos(x9)) * x6
        ) + w2
        dx3 = (
            np.sin(x8) * x4
            - np.sin(x7) * np.cos(x8) * x5
            - np.cos(x7) * np.cos(x8) * x6
            + w3
        )
        dx4 = x12 * x5 - x11 * x6 - self.gravitational_acceleration * np.sin(x8) + w4
        dx5 = (
            x10 * x6
            - x12 * x4
            + self.gravitational_acceleration * np.cos(x8) * np.sin(x7)
            + w5
        )
        dx6 = (
            x11 * x4
            - x10 * x5
            + self.gravitational_acceleration * np.cos(x8) * np.cos(x7)
            - (F / self.total_mass)
        ) + w6
        dx7 = x10 + np.sin(x7) * np.tan(x8) * x11 + np.cos(x7) * np.tan(x8) * x12 + w7
        dx8 = np.cos(x7) * x11 - np.sin(x7) * x12 + w8
        dx9 = (np.sin(x7) / np.cos(x8)) * x11 + (np.cos(x7) / np.cos(x8)) * x12 + w9
        dx10 = (
            ((self.Jy - self.Jz) / self.Jx) * x11 * x12 + (1 / self.Jx) * tau_phi + w10
        )
        dx11 = (
            ((self.Jz - self.Jx) / self.Jy) * x10 * x12 + (1 / self.Jy) * tau_psi + w11
        )
        dx12 = (
            ((self.Jx - self.Jy) / self.Jz) * x10 * x11
            + (1 / self.Jz) * tau_theta
            + w12
        )

        return np.array(
            [dx1, dx2, dx3, dx4, dx5, dx6, dx7, dx8, dx9, dx10, dx11, dx12],
            dtype=np.float32,
        )



"""Satellite rendezvous and docking.

This example demonstrates the optimal controller synthesis algorithm on a satellite rendezvous and docking problem with CWH dynamics.

Example:
    To run the example, use the following command:

        $ python -m examples.benchmark_cwh_problem

.. [1] `Stochastic Optimal Control via
        Hilbert Space Embeddings of Distributions, 2021
        Adam J. Thorpe, Meeko M. K. Oishi
        IEEE Conference on Decision and Control,
        <https://arxiv.org/abs/2103.12759>`_

"""


import gym
import gym_socks

import logging

import numpy as np
from numpy.linalg import norm

from sacred import Experiment

from functools import partial
from sklearn.metrics.pairwise import rbf_kernel

from gym_socks.algorithms.control import KernelControlFwd
from gym_socks.algorithms.control import KernelControlBwd

from gym_socks.envs.sample import transpose_sample

from examples._computation_timer import ComputationTimer

from examples.ingredients.system_ingredient import system_ingredient
from examples.ingredients.system_ingredient import set_system_seed
from examples.ingredients.system_ingredient import make_system

from examples.ingredients.simulation_ingredient import simulation_ingredient
from examples.ingredients.simulation_ingredient import simulate_system

from examples.ingredients.sample_ingredient import sample_ingredient
from examples.ingredients.sample_ingredient import generate_sample
from examples.ingredients.sample_ingredient import generate_admissible_actions

from examples.ingredients.cwh_ingredient import cwh_ingredient
from examples.ingredients.cwh_ingredient import make_cost
from examples.ingredients.cwh_ingredient import make_constraint

from examples.ingredients.plotting_ingredient import plotting_ingredient
from examples.ingredients.plotting_ingredient import update_rc_params


@system_ingredient.config
def system_config():
    system_id = "CWH4DEnv-v0"

    sampling_time = 20
    time_horizon = 100

    action_space = {
        "lower_bound": -0.1,
        "upper_bound": 0.1,
    }


@sample_ingredient.config
def sample_config():

    sample_space = {
        "sample_scheme": "grid",
        "lower_bound": [-1.1, -1.1, -0.06, -0.06],
        "upper_bound": [1.1, 0.1, 0.06, 0.06],
        "grid_resolution": [10, 10, 5, 5],
    }

    sample_policy = {
        "sample_scheme": "random",
    }

    action_space = {
        "sample_scheme": "uniform",
        "lower_bound": -0.05,
        "upper_bound": 0.05,
        "sample_size": 500,
    }


@simulation_ingredient.config
def simulation_config():

    initial_condition = [-0.75, -0.75, 0, 0]


ex = Experiment(
    ingredients=[
        system_ingredient,
        simulation_ingredient,
        sample_ingredient,
        cwh_ingredient,
        plotting_ingredient,
    ]
)


@ex.config
def config(sample):
    """Experiment configuration variables.

    SOCKS uses sacred to run experiments in order to ensure repeatability. Configuration
    variables are parameters that are passed to the experiment, such as the random seed,
    and can be specified at the command-line.

    Example:
        To run the experiment normally, use:

            $ python -m <experiment>

        The full configuration can be viewed using:

            $ python -m <experiment> print_config

        To specify configuration variables, use `with variable=value`, e.g.

            $ python -m <experiment> with seed=123 system.time_horizon=5

    .. _sacred:
        https://sacred.readthedocs.io/en/stable/index.html

    """

    sigma = 0.35  # Kernel bandwidth parameter.
    # Regularization parameter.
    # regularization_param = 1 / (sample["sample_space"]["sample_size"] ** 2)
    regularization_param = 1e-5

    # Whether or not to use dynamic programming (backward in time) algorithm.
    dynamic_programming = False
    batch_size = None  # (Optional) batch size for batch computation.
    heuristic = False  # Whether to use the heuristic solution.

    verbose = True

    results_filename = "results/data.npy"
    no_plot = False


@ex.main
def main(
    seed,
    sigma,
    regularization_param,
    dynamic_programming,
    batch_size,
    heuristic,
    verbose,
    results_filename,
    no_plot,
    simulation,
    _log,
):
    """Main experiment."""

    # Make the system.
    env = make_system()

    # Set the random seed.
    set_system_seed(seed=seed, env=env)

    # # Generate the sample.
    S = generate_sample(seed=seed, env=env)

    # Generate the set of admissible control actions.
    A = generate_admissible_actions(seed=seed, env=env)

    # Compute the target trajectory.
    cost_fn = make_cost(env=env)
    constraint_fn = make_constraint(env=env)

    if dynamic_programming is True:
        alg_class = KernelControlBwd
    else:
        alg_class = KernelControlFwd

    with ComputationTimer():

        # Compute policy.
        policy = alg_class(
            num_steps=env.num_time_steps,
            cost_fn=cost_fn,
            constraint_fn=constraint_fn,
            heuristic=heuristic,
            verbose=verbose,
            kernel_fn=partial(rbf_kernel, gamma=1 / (2 * (sigma ** 2))),
            regularization_param=regularization_param,
            batch_size=batch_size,
        )

        policy.train(S=S, A=A)

    trajectory = simulate_system(env, policy)

    with open(results_filename, "wb") as f:
        np.save(f, trajectory)

    if not no_plot:
        plot_results()


@plotting_ingredient.config_hook
def plot_config(config, command_name, logger):
    if command_name in {"main", "plot_results", "plot_sample"}:
        return {
            "target_trajectory_style": {
                "lines.marker": "x",
                "lines.linestyle": "--",
                "lines.color": "C0",
            },
            "trajectory_style": {
                "lines.linewidth": 1,
                "lines.linestyle": "-",
                "lines.color": "C1",
            },
            "axes": {
                "xlabel": r"$x_1$",
                "ylabel": r"$x_2$",
                "xlim": (-1.1, 1.1),
                "ylim": (-1.1, 0.1),
            },
        }


@ex.command(unobserved=True)
def plot_results(plot_cfg, _log):
    """Plot the results of the experiement."""

    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    logging.getLogger("PIL").setLevel(logging.WARNING)

    # Dynamically load for speed.
    import matplotlib

    matplotlib.use("Agg")
    update_rc_params(matplotlib, plot_cfg["rc_params"])

    import matplotlib.pyplot as plt

    with open("results/data.npy", "rb") as f:
        trajectory = np.load(f)

    fig = plt.figure()
    ax = fig.add_subplot(111, **plot_cfg["axes"])

    # plot constraint box
    verts = [(-1, -1), (1, -1), (0, 0), (-1, -1)]
    codes = [
        matplotlib.path.Path.MOVETO,
        matplotlib.path.Path.LINETO,
        matplotlib.path.Path.LINETO,
        matplotlib.path.Path.CLOSEPOLY,
    ]

    path = matplotlib.path.Path(verts, codes)
    plt.gca().add_patch(matplotlib.patches.PathPatch(path, fc="none", ec="blue"))

    plt.gca().add_patch(plt.Rectangle((-0.2, -0.2), 0.4, 0.2, fc="none", ec="green"))

    # Plot generated trajectory.
    with plt.style.context(plot_cfg["trajectory_style"]):
        trajectory = np.array(trajectory, dtype=np.float32)
        plt.plot(
            trajectory[:, 0],
            trajectory[:, 1],
            label="System Trajectory",
        )

    plt.legend()

    plt.savefig(plot_cfg["plot_filename"])


@ex.command(unobserved=True)
def plot_sample(seed, plot_cfg):
    """Plot a sample taken from the system."""

    import matplotlib

    matplotlib.use("Agg")
    update_rc_params(matplotlib, plot_cfg["rc_params"])

    import matplotlib.pyplot as plt

    matplotlib.set_loglevel("notset")
    plt.set_loglevel("notset")

    # Make the system.
    env = make_system()

    # Set the random seed.
    set_system_seed(seed=seed, env=env)

    # # Generate the sample.
    S = generate_sample(seed=seed, env=env)
    X, U, Y = transpose_sample(S)
    X = np.array(X)
    U = np.array(U)
    Y = np.array(Y)

    fig = plt.figure()
    ax = fig.add_subplot(111, **plot_cfg["axes"])

    # plot constraint box
    verts = [(-1, -1), (1, -1), (0, 0), (-1, -1)]
    codes = [
        matplotlib.path.Path.MOVETO,
        matplotlib.path.Path.LINETO,
        matplotlib.path.Path.LINETO,
        matplotlib.path.Path.CLOSEPOLY,
    ]

    path = matplotlib.path.Path(verts, codes)
    plt.gca().add_patch(matplotlib.patches.PathPatch(path, fc="none", ec="blue"))

    plt.gca().add_patch(plt.Rectangle((-0.2, -0.2), 0.4, 0.2, fc="none", ec="green"))

    plt.scatter(X[:, 0], X[:, 1], marker=".")
    plt.scatter(Y[:, 0], Y[:, 1], marker=".")

    plt.savefig("results/plot_sample.png")


if __name__ == "__main__":
    ex.run_commandline()



"""Stochastic reachability.

This example shows the stochastic reachability algorithm.

By default, the system is a double integrator (2D stochastic chain of integrators).

Example:
    To run the example, use the following command:

        $ python -m examples.benchmark_stochastic_reachability

.. [1] `Model-Free Stochastic Reachability
        Using Kernel Distribution Embeddings, 2019
        Adam J. Thorpe, Meeko M. K. Oishi
        IEEE Control Systems Letters,
        <https://arxiv.org/abs/1908.00697>`_

"""

import gym
import gym_socks

import logging

import numpy as np

from sacred import Experiment

from functools import partial
from sklearn.metrics.pairwise import rbf_kernel

from gym_socks.algorithms.reach.kernel_sr import kernel_sr

from examples._computation_timer import ComputationTimer

from examples.ingredients.system_ingredient import system_ingredient
from examples.ingredients.system_ingredient import set_system_seed
from examples.ingredients.system_ingredient import make_system

from examples.ingredients.sample_ingredient import sample_ingredient
from examples.ingredients.sample_ingredient import generate_sample

from examples.ingredients.backward_reach_ingredient import backward_reach_ingredient
from examples.ingredients.backward_reach_ingredient import generate_test_points
from examples.ingredients.backward_reach_ingredient import compute_test_point_ranges
from examples.ingredients.backward_reach_ingredient import generate_tube

from examples.ingredients.plotting_ingredient import plotting_ingredient
from examples.ingredients.plotting_ingredient import update_rc_params


@system_ingredient.config
def system_config():
    system_id = "2DIntegratorEnv-v0"

    time_horizon = 4
    sampling_time = 0.25


@sample_ingredient.config
def sample_config():

    sample_space = {
        "sample_scheme": "grid",
        "lower_bound": -1.1,
        "upper_bound": 1.1,
        "grid_resolution": 50,
    }

    sample_policy = {"sample_scheme": "zero"}

    action_space = {"sample_scheme": "grid"}


@backward_reach_ingredient.config
def backward_reach_config():

    # By default, the constraint tube is defined as a box [-1, 1]^d.
    constraint_tube_bounds = {"lower_bound": -1, "upper_bound": 1}
    # By default, the target tube is defined as a box [-0.5, 0.5]^d.
    target_tube_bounds = {"lower_bound": -0.5, "upper_bound": 0.5}

    test_points = {
        "lower_bound": -1,
        "upper_bound": 1,
        "grid_resolution": 50,
    }


ex = Experiment(
    ingredients=[
        system_ingredient,
        sample_ingredient,
        backward_reach_ingredient,
        plotting_ingredient,
    ]
)


@ex.config
def config():
    """Experiment configuration variables.

    SOCKS uses sacred to run experiments in order to ensure repeatability. Configuration
    variables are parameters that are passed to the experiment, such as the random seed,
    and can be specified at the command-line.

    Example:
        To run the experiment normally, use:

            $ python -m <experiment>

        The full configuration can be viewed using:

            $ python -m <experiment> print_config

        To specify configuration variables, use `with variable=value`, e.g.

            $ python -m <experiment> with seed=123 system.time_horizon=5

    .. _sacred:
        https://sacred.readthedocs.io/en/stable/index.html

    """

    sigma = 0.1
    regularization_param = 1

    batch_size = None

    verbose = True

    results_filename = "results/data.npy"
    no_plot = False


@ex.main
def main(
    seed,
    _log,
    sigma,
    regularization_param,
    backward_reach,
    batch_size,
    verbose,
    results_filename,
    no_plot,
):
    """Main experiment."""

    env = make_system()

    # Set the random seed.
    set_system_seed(seed, env)

    # Generate the target and constraint tubes.
    target_tube = generate_tube(env, backward_reach["target_tube_bounds"])
    constraint_tube = generate_tube(env, backward_reach["constraint_tube_bounds"])

    # Generate the sample.
    S = generate_sample(seed=seed, env=env)

    # Generate the test points.
    T = generate_test_points(env=env)

    with ComputationTimer():

        safety_probabilities = kernel_sr(
            S=S,
            T=T,
            num_steps=env.num_time_steps,
            constraint_tube=constraint_tube,
            target_tube=target_tube,
            problem=backward_reach["problem"],
            regularization_param=regularization_param,
            kernel_fn=partial(rbf_kernel, gamma=1 / (2 * (sigma ** 2))),
            batch_size=batch_size,
            verbose=verbose,
        )

    # Save the result to NPY file.
    _log.debug(f"Saving the results to file {results_filename}.")
    xi = compute_test_point_ranges(env)
    with open(results_filename, "wb") as f:
        np.save(f, xi)
        np.save(f, safety_probabilities)

    if not no_plot:
        plot_results()


@plotting_ingredient.config_hook
def _plot_config(config, command_name, logger):
    if command_name in {"main", "plot_results"}:
        return {
            "plot_time": 0,
            "plot_system_dims": [0, 1],
            "pcolor_style": {
                "vmin": 0,
                "vmax": 1,
                "shading": "auto",
            },
            "axes": {
                "xlabel": r"$x_1$",
                "ylabel": r"$x_2$",
            },
            "colorbar": True,
        }


@ex.command(unobserved=True)
def plot_results(
    plot_cfg,
    results_filename,
):
    """Plot the results of the experiement."""

    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    logging.getLogger("PIL").setLevel(logging.WARNING)

    # Dynamically load for speed.
    import matplotlib

    matplotlib.use("Agg")
    update_rc_params(matplotlib, plot_cfg["rc_params"])

    import matplotlib.pyplot as plt

    # Load the result from NPY file.
    with open(results_filename, "rb") as f:
        xi = np.load(f)
        safety_probabilities = np.load(f)

    plot_system_dims = plot_cfg["plot_system_dims"]
    x1 = xi[plot_system_dims[0]]
    x2 = xi[plot_system_dims[1]]
    XX, YY = np.meshgrid(x1, x2, indexing="ij")
    Z = safety_probabilities[plot_cfg["plot_time"]].reshape(XX.shape)

    # Plot flat color map.
    fig = plt.figure()
    ax = plt.axes(**plot_cfg["axes"])

    plt.pcolor(XX, YY, Z, **plot_cfg["pcolor_style"])

    if plot_cfg["colorbar"] is True:
        plt.colorbar()

    plt.savefig(plot_cfg["plot_filename"])


@plotting_ingredient.config_hook
def plot_config_3d(config, command_name, logger):
    if command_name in {"plot_results_3d"}:
        return {
            "plot_time": 0,
            "plot_system_dims": [0, 1],
            "axes": {
                "elev": 30,
                "azim": -45,
                "xlabel": r"$x_1$",
                "ylabel": r"$x_2$",
                "zlabel": r"$\Pr$",
                "zlim": (0, 1),
            },
            "cmap": "viridis",
            "surface_style": {
                "linewidth": 0,
                "antialiased": False,
            },
        }


@ex.command(unobserved=True)
def plot_results_3d(plot_cfg, results_filename):

    # Dynamically load for speed.
    import matplotlib

    matplotlib.use("Agg")
    rc_params = matplotlib.rc_params_from_file(
        fname=plot_cfg["rc_params_filename"],
        use_default_template=True,
    )
    matplotlib.rcParams.update(rc_params)

    import matplotlib.pyplot as plt
    import matplotlib.cm as cm

    matplotlib.set_loglevel("notset")
    plt.set_loglevel("notset")

    # Load the result from NPY file.
    with open(results_filename, "rb") as f:
        xi = np.load(f)
        safety_probabilities = np.load(f)

    plot_system_dims = plot_cfg["plot_system_dims"]
    x1 = xi[plot_system_dims[0]]
    x2 = xi[plot_system_dims[1]]
    XX, YY = np.meshgrid(x1, x2, indexing="ij")
    Z = safety_probabilities[plot_cfg["plot_time"]].reshape(XX.shape)

    # Plot 3D projection.
    fig = plt.figure()
    ax = plt.axes(projection="3d", **plot_cfg["axes"])

    mappable = cm.ScalarMappable(cmap=plot_cfg["cmap"])
    mappable.set_array(Z)
    mappable.set_clim(vmin=0, vmax=1)

    surf = ax.plot_surface(
        XX,
        YY,
        Z,
        cmap=mappable.cmap,
        norm=mappable.norm,
        **plot_cfg["surface_style"],
    )

    plt.savefig(plot_cfg["plot_filename"])


if __name__ == "__main__":
    ex.run_commandline()



import unittest
from unittest import mock
from unittest.mock import patch

import gym

from gym_socks.envs import NDIntegratorEnv

import numpy as np

from examples.ingredients.sample_ingredient import sample_ingredient
from examples.ingredients.sample_ingredient import generate_sample
from examples.ingredients.sample_ingredient import generate_admissible_actions


class TestSampleIngredient(unittest.TestCase):
    def test_sample_config(cls):
        """Test the ingredient config."""
        ingredient = sample_ingredient
        cls.assertEqual(len(ingredient.configurations), 1)


class TestGenerateSample(unittest.TestCase):
    def test_generate_sample(cls):
        """Test that generate_sample produces a sample from a system."""

        pass


class TestGenerateAdmissibleActions(unittest.TestCase):
    def test_generate_admissible_actions(cls):
        """Test that generate_admissible_actions produces a sample from a system."""

        pass



from gym_socks.utils import indicator_fn


def _fht_step(Y, V, constraint_set, target_set):
    """First-hitting time problem backward recursion step.

    This function implements the backward recursion step for the first-hitting time
    problem, given by::

        V_t(x) = 1[T](x) + 1[K\T](x) E[V_t+1]

    Args:
        Y: Vector of data points in the sample.
        V: Vector of coefficients from the kernel-based approximation.
        constraint_set: A space or function which determines whether a point is in the
            constraint set. This can be either a space from gym.spaces or a function
            which acts as a 0/1 classifier. See the `indicator_fn` implementation for
            more information.
        target_set: A space or function which determines whether a point is in the
            target set. This can be either a space from gym.spaces or a function which
            acts as a 0/1 classifier. See the `indicator_fn` implementation for more
            information.

    Returns:
        The kernel-based approximation of the value function.

    """

    Y_in_constraint_set = indicator_fn(Y, constraint_set)
    Y_in_target_set = indicator_fn(Y, target_set)

    return Y_in_target_set + (Y_in_constraint_set & ~Y_in_target_set) * V


def _tht_step(Y, V, constraint_set, target_set):
    """Terminal-hitting time problem backward recursion step.

    This function implements the backward recursion step for the terminal-hitting time
    problem, given by::

        V_t(x) = 1[K](x) E[V_t+1]

    Args:
        Y: Vector of data points in the sample.
        V: Vector of coefficients from the kernel-based approximation.
        constraint_set: A space or function which determines whether a point is in the
            constraint set. This can be either a space from gym.spaces or a function
            which acts as a 0/1 classifier. See the `indicator_fn` implementation for
            more information.
        target_set: This argument is unused in the terminal-hitting time problem. The
            reson it is allowed is to avoid extra if statements in the implementation.

    Returns:
        The kernel-based approximation of the value function.

    """

    Y_in_constraint_set = indicator_fn(Y, constraint_set)

    return Y_in_constraint_set * V



import unittest
from unittest import mock
from unittest.mock import patch

import gym

import gym_socks.envs

import numpy as np

from scipy.constants import gravitational_constant


class Test4DCWHSystem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.env = gym_socks.envs.cwh.CWH4DEnv()

    def test_set_parameters(cls):

        state_matrix = cls.env.state_matrix
        input_matrix = cls.env.input_matrix

        # system parameters
        cls.env.orbital_radius = 850 + 6378.1
        cls.env.gravitational_constant = gravitational_constant
        cls.env.celestial_body_mass = 5.9472e24
        cls.env.chief_mass = 300

        cls.assertTrue(np.allclose(state_matrix, cls.env.state_matrix))
        cls.assertTrue(np.allclose(input_matrix, cls.env.input_matrix))

    @patch.object(gym_socks.envs.cwh.CWH4DEnv, "generate_disturbance")
    def test_known_trajectory(cls, mock_generate_disturbance):
        """Test against specific known trajectory. Sanity check."""
        mock_generate_disturbance.return_value = np.zeros((4,))

        env = cls.env

        env.state = np.array([-0.1, -0.1, 0, 0], dtype=np.float32)
        action = np.array([0, 0], dtype=np.float32)

        trajectory = []
        trajectory.append(env.state)

        for i in range(3):
            obs, cost, done, _ = env.step(action)
            trajectory.append(obs)

        trajectory = np.array(trajectory)

        n = env.angular_velocity
        nt = n * env.sampling_time
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        A = [
            [4 - 3 * cos_nt, 0, (1 / n) * sin_nt, (2 / n) * (1 - cos_nt)],
            [
                6 * (sin_nt - nt),
                1,
                -(2 / n) * (1 - cos_nt),
                (1 / n) * (4 * sin_nt - 3 * nt),
            ],
            [3 * n * sin_nt, 0, cos_nt, 2 * sin_nt],
            [-6 * n * (1 - cos_nt), 0, -2 * sin_nt, 4 * cos_nt - 3],
        ]

        state = np.array([-0.1, -0.1, 0, 0])

        groundTruth = []
        groundTruth.append(state)

        for i in range(3):
            state = np.matmul(A, state.T)
            groundTruth.append(state)

        # tests if the two arrays are equivalent, within tolerance
        cls.assertTrue(
            np.allclose(trajectory, groundTruth),
            "Generated trajectory should match known trajectory generated using dynamics.",
        )

        state = np.array([-0.1, 0.1, 0, 0])

        falseTrajectory = []
        falseTrajectory.append(state)

        for i in range(3):
            state = np.matmul(A, state.T)
            falseTrajectory.append(state)

        cls.assertFalse(
            np.allclose(trajectory, falseTrajectory),
            "Generated trajectory should not match known false trajectory generated using dynamics.",
        )


class Test6DCWHSystem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.env = gym_socks.envs.cwh.CWH6DEnv()

    @patch.object(gym_socks.envs.cwh.CWH6DEnv, "generate_disturbance")
    def test_known_trajectory(cls, mock_generate_disturbance):
        """Test against specific known trajectory. Sanity check."""
        mock_generate_disturbance.return_value = np.zeros((6,))

        env = cls.env

        env.state = np.array([-0.1, -0.1, 0, 0, 0, 0], dtype=np.float32)
        action = np.array([0, 0, 0], dtype=np.float32)

        trajectory = []
        trajectory.append(env.state)

        for i in range(3):
            obs, cost, done, _ = env.step(action)
            trajectory.append(obs)

        trajectory = np.array(trajectory)

        n = env.angular_velocity
        nt = n * env.sampling_time
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        A = [
            [4 - 3 * cos_nt, 0, 0, (1 / n) * sin_nt, (2 / n) * (1 - cos_nt), 0],
            [
                6 * (sin_nt - nt),
                1,
                0,
                -(2 / n) * (1 - cos_nt),
                (1 / n) * (4 * sin_nt - 3 * nt),
                0,
            ],
            [0, 0, cos_nt, 0, 0, (1 / n) * sin_nt],
            [3 * n * sin_nt, 0, 0, cos_nt, 2 * sin_nt, 0],
            [-6 * n * (1 - cos_nt), 0, 0, -2 * sin_nt, 4 * cos_nt - 3, 0],
            [0, 0, -n * sin_nt, 0, 0, cos_nt],
        ]

        state = np.array([-0.1, -0.1, 0, 0, 0, 0])

        groundTruth = []
        groundTruth.append(state)

        for i in range(3):
            state = np.matmul(A, state.T)
            groundTruth.append(state)

        # tests if the two arrays are equivalent, within tolerance
        cls.assertTrue(
            np.allclose(trajectory, groundTruth),
            "Generated trajectory should match known trajectory generated using dynamics.",
        )

        state = np.array([-0.1, 0.1, 0, 0, 0, 0])

        falseTrajectory = []
        falseTrajectory.append(state)

        for i in range(3):
            state = np.matmul(A, state.T)
            falseTrajectory.append(state)

        cls.assertFalse(
            np.allclose(trajectory, falseTrajectory),
            "Generated trajectory should not match known false trajectory generated using dynamics.",
        )



__all__ = [
    "backward_reach_ingredient",
    "sample_ingredient",
    "simulation_ingredient",
    "system_ingredient",
    "tracking_ingredient",
]

from examples.ingredients.backward_reach_ingredient import backward_reach_ingredient
from examples.ingredients.backward_reach_ingredient import generate_tube

from examples.ingredients.sample_ingredient import sample_ingredient
from examples.ingredients.simulation_ingredient import simulation_ingredient
from examples.ingredients.system_ingredient import system_ingredient
from examples.ingredients.tracking_ingredient import tracking_ingredient



import unittest
from unittest.mock import patch

import gym

import gym_socks.envs

import numpy as np

system_list = [
    gym_socks.envs.NDIntegratorEnv(1),
    gym_socks.envs.NDIntegratorEnv(2),
    gym_socks.envs.NDIntegratorEnv(3),
    gym_socks.envs.NDIntegratorEnv(4),
    gym_socks.envs.NDPointMassEnv(1),
    gym_socks.envs.NDPointMassEnv(2),
    gym_socks.envs.NDPointMassEnv(3),
    gym_socks.envs.NDPointMassEnv(4),
    gym_socks.envs.NonholonomicVehicleEnv(),
    gym_socks.envs.PlanarQuadrotorEnv(),
    gym_socks.envs.CWH4DEnv(),
    gym_socks.envs.CWH6DEnv(),
    gym_socks.envs.QuadrotorEnv(),
    gym_socks.envs.TORAEnv(),
]


class TestEnvironmentsRun(unittest.TestCase):
    def test_envs_run(cls):
        """Assert that environments can run."""

        for env in system_list:
            with cls.subTest(msg=f"Testing with {type(env)}."):

                try:
                    obs = env.reset()

                    for i in range(env.num_time_steps):

                        # get action
                        action = env.action_space.sample()

                        # apply action
                        obs, cost, done, _ = env.step(action)

                except Exception as e:
                    cls.fail(f"Simulating system {type(env)} raised an exception.")


class TestDynamicalSystem(unittest.TestCase):
    """Dynamical system tests."""

    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def setUp(cls):
        cls.env = gym_socks.envs.dynamical_system.DynamicalSystem(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            seed=1,
        )

    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def test_should_fail_without_spaces(cls):
        """Test that system throws an error if spaces are not specified."""
        with cls.assertRaises(ValueError):
            system = gym_socks.envs.dynamical_system.DynamicalSystem(
                observation_space=None, action_space=None
            )

        with cls.assertRaises(ValueError):
            system = gym_socks.envs.dynamical_system.DynamicalSystem(
                observation_space=gym.spaces.Box(
                    low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
                ),
                action_space=None,
            )

        with cls.assertRaises(ValueError):
            system = gym_socks.envs.dynamical_system.DynamicalSystem(
                observation_space=None,
                action_space=gym.spaces.Box(
                    low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
                ),
            )

    # def test_system_warns_time_step(cls):
    #     """
    #     Test that system throws a warning if sampling time is greater than time horizon.
    #     """

    #     with cls.assertWarns(Warning):
    #         cls.env.time_horizon = 1
    #         cls.env.sampling_time = 2

    #     with cls.assertWarns(Warning):
    #         cls.env.sampling_time = 1
    #         cls.env.time_horizon = 0.1

    def test_system_num_time_steps(cls):
        """System returns correct number of time steps."""

        cls.env.time_horizon = 1
        cls.env.sampling_time = 0.1

        # Note that this is not an error, and has to do with floating point precision.
        # See: https://docs.python.org/3/tutorial/floatingpoint.html
        cls.assertEqual(cls.env.num_time_steps, 9)
        cls.assertEqual(cls.env.time_horizon, 1)
        cls.assertEqual(cls.env.sampling_time, 0.1)

        cls.env.time_horizon = 10.0
        cls.env.sampling_time = 1.0

        # Note that this is not an error, and has to do with floating point precision.
        # See: https://docs.python.org/3/tutorial/floatingpoint.html
        cls.assertEqual(cls.env.num_time_steps, 10)
        cls.assertEqual(cls.env.time_horizon, 10)
        cls.assertEqual(cls.env.sampling_time, 1.0)

        cls.env.time_horizon = 5
        cls.env.sampling_time = 1

        cls.assertEqual(cls.env.num_time_steps, 5)
        cls.assertEqual(cls.env.time_horizon, 5)
        cls.assertEqual(cls.env.sampling_time, 1)

    def test_dims(cls):
        """State and action dims should match spaces."""
        cls.assertEqual(cls.env.state_dim, (1,))
        cls.assertEqual(cls.env.action_dim, (1,))

        cls.env.observation_space = gym.spaces.Box(
            low=-1, high=1, shape=(1,), dtype=np.float32
        )

        cls.assertEqual(cls.env.state_dim, (1,))

    def test_cannot_change_dims(cls):
        """Cannot change the state and action dims."""

        with cls.assertRaises(AttributeError):
            cls.env.state_dim = 1

        with cls.assertRaises(AttributeError):
            cls.env.action_dim = 1

    def test_reset_returns_valid_state(cls):
        """Reset should return a valid state."""
        cls.env.reset()
        cls.assertTrue(cls.env.observation_space.contains(cls.env.state))

    def test_default_dynamics_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            state = cls.env.reset()
            action = cls.env.action_space.sample()
            disturbance = cls.env.generate_disturbance(0, state, action)
            cls.env.dynamics(0, state, action, disturbance)

    def test_default_close_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            cls.env.close()

    def test_default_render_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            cls.env.render()



from abc import ABC, abstractmethod

import gym
from gym.utils import seeding

import gym_socks

import numpy as np
from scipy.integrate import solve_ivp


class DynamicalSystem(gym.Env, ABC):
    """Base class for dynamical system models.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from `DynamicalSystem`, and define a
    custom `dynamics` function.

    Example:

        >>> from gym_socks.envs.dynamical_system import DynamicalSystem
        >>> class CustomDynamicalSystem(DynamicalSystem):
        ...     def __init__(self):
        ...         super().__init__(state_dim=2, action_dim=1)
        ...
        ...     def dynamics(self, t, x, u, w):
        ...         return u
        ...
        ...     def reset(self):
        ...         self.state = self.np_random.uniform(
        ...             low=-1, high=1, size=self.observation_space.shape
        ...         )
        ...         return np.array(self.state)

    Note:
        * The state space and input space are assumed to be R^n and R^m, where n and m
        are set by state_dim and action_dim above (though these can be altered, see
        gym.spaces for more info).
        * The dynamics function (defined by you) returns dx/dt, and the system is
        integrated using scipy.integrate.solve_ivp to determine the state at the next
        time instant and discretize the system.
        * The reset function sets a new initial condition for the system.

    The system can then be simulated using the standard gym environment.

    Example:

        >>> import gym
        >>> import numpy as np
        >>> env = CustomDynamicalSystem()
        >>> env.reset()
        >>> num_steps = env.num_time_steps
        >>> for i in range(num_steps):
        ...     action = env.action_space.sample()
        ...     obs, reward, done, _ = env.step(action)
        >>> env.close()

    Args:
        observation_space: The space of system observations.
        state_space: The state space of the system.
        action_space: The action (input) space of the system.
        seed: Random seed.
        euler: Whether to use the Euler approximation for simulation.

    """

    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        observation_space: gym.Space = None,
        state_space: gym.Space = None,
        action_space: gym.Space = None,
        seed: int = None,
        euler: bool = False,
        *args,
        **kwargs,
    ):

        if observation_space is None:
            raise ValueError("Must supply an observation_space.")

        self.observation_space = observation_space

        if state_space is None:
            state_space = observation_space

        self.state_space = state_space

        if action_space is None:
            raise ValueError("Must supply an action_space.")

        self.action_space = action_space

        # Time parameters.
        self._time_horizon = 1
        self._sampling_time = 0.1

        self.state = None

        self._seed = None
        self._np_random = None
        if seed is not None:
            self._seed = self.seed(seed)

        self._euler = euler

    @property
    def np_random(self):
        """Random number generator."""
        if self._np_random is None:
            self._seed = self.seed()

        return self._np_random

    def seed(self, seed=None):
        """Sets the seed of the random number generator.

        Args:
            seed: Integer value representing the random seed.

        Returns:
            The seed of the RNG.

        """
        self._np_random, seed = seeding.np_random(seed)
        return [seed]

    @property
    def sampling_time(self):
        """Sampling time of the system."""
        return self._sampling_time

    @sampling_time.setter
    def sampling_time(self, value):
        # msg = f"Sampling time {value} is less than time horizon {self._time_horizon}."
        # if value > self._time_horizon:
        #     gym_socks.logging.warn(msg)
        self._sampling_time = value

    @property
    def time_horizon(self):
        """Time horizon of the system."""
        return self._time_horizon

    @time_horizon.setter
    def time_horizon(self, value):
        # msg = f"Sampling time {value} is less than time horizon {self._time_horizon}."
        # if value < self._sampling_time:
        #     gym_socks.logging.warn(msg)
        self._time_horizon = value

    @property
    def num_time_steps(self):
        """Number of time steps.

        Defined as the `time_horizon // sampling_time`.

        """
        return int(self._time_horizon // self._sampling_time)

    @property
    def observation_dim(self):
        """Dimensionality of the observation space."""
        return self.observation_space.shape

    @property
    def state_dim(self):
        """Dimensionality of the state space."""
        return self.state_space.shape

    @property
    def action_dim(self):
        """Dimensionality of the action space."""
        return self.action_space.shape

    def step(self, action, time=0):
        """Step function defined by OpenAI Gym.

        Advances the system forward one time step.

        Args:
            action: Action (input) applied to the system at the current time step.
            time: Time of the simulation. Used primarily for time-varying systems.

        Returns:
            obs: The observation vector. Generally, it is the state of the system
                corrupted by some measurement noise. If the system is fully observable,
                this is the state of the system at the next time step.
            cost: The cost (reward) obtained by the system for taking action u in state
                x and transitioning to state y. In general, this is not typically used with `DynamicalSystem` models.
            done: Flag to indicate the simulation has terminated. Usually toggled by
                guard conditions, which terminates the simulation if the system
                violates certain operating constraints.
            info: Extra information.

        """

        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        # Generate disturbance.
        disturbance = self.generate_disturbance(time, self.state, action)

        if self._euler is True:
            next_state = self.state + self.sampling_time * self.dynamics(
                time, self.state, action, disturbance
            )
            self.state = next_state
        else:
            # Solve the initial value problem.
            sol = solve_ivp(
                self.dynamics,
                [time, time + self.sampling_time],
                self.state,
                args=(
                    action,
                    disturbance,
                ),
            )
            *_, self.state = sol.y.T

        # Generate observation.
        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def reset(self):
        """Reset the system to a random initial condition."""
        self.state = self.state_space.sample()

    def render(self, mode="human"):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def generate_disturbance(self, time, state, action):
        """Generate disturbance."""
        return self._np_random.standard_normal(size=self.state_space.shape)

    @abstractmethod
    def dynamics(self, time, state, action, disturbance):
        """Dynamics for the system.

        The dynamics are typically specified by a function::
            y = f(t, x, u, w)
                  ┬  ┬  ┬  ┬
                  │  │  │  └┤ w : Disturbance
                  │  │  └───┤ u : Control action
                  │  └──────┤ x : System state
                  └─────────┤ t : Time variable

        """
        raise NotImplementedError

    def generate_observation(self, time, state, action):
        """Generate observation."""
        return state

    def cost(self, time, state, action):
        """Cost function for the system."""
        return 0.0



import unittest
from unittest.mock import patch

import gym

import gym_socks.envs

import numpy as np


class TestNonholonomicSystem(unittest.TestCase):
    def test_corrects_angle(cls):
        system = gym_socks.envs.NonholonomicVehicleEnv()
        system.disturbance_space = gym.spaces.Box(
            low=0, high=0, shape=(3,), dtype=np.float32
        )

        system.state = np.array([0, 0, 4 * np.pi])
        action = np.array([0, 0], dtype=np.float32)
        obs, cost, done, _ = system.step(action)

        cls.assertTrue((system.state[2] <= 2 * np.pi) & (system.state[2] >= -2 * np.pi))



"""Kernel-based stochastic reachability.

References:
    .. [1] `Model-Free Stochastic Reachability
            Using Kernel Distribution Embeddings, 2019
           Adam J. Thorpe, Meeko M. K. Oishi
           IEEE Control Systems Letters,
           <https://arxiv.org/abs/1908.00697>`_

"""

from functools import partial

import gym_socks
from gym_socks.algorithms.algorithm import AlgorithmInterface
from gym_socks.algorithms.reach.reach_common import _tht_step, _fht_step
from gym_socks.envs.sample import transpose_sample
from gym_socks.kernel.metrics import rbf_kernel, regularized_inverse

from gym_socks.utils import normalize, indicator_fn, generate_batches
from gym_socks.utils.logging import ms_tqdm, _progress_fmt
from tqdm.contrib.logging import logging_redirect_tqdm

import numpy as np


def _compute_backward_recursion(
    Y,
    W,
    CXY,
    CUA,
    num_steps=None,
    constraint_tube=None,
    target_tube=None,
    value_functions=None,
    out=None,
    step_fn=None,
    verbose=False,
):
    # Compute beta.
    gym_socks.logger.debug("Computing beta.")
    betaXY = normalize(
        np.einsum("ii,ij,ik->ijk", W, CXY, CUA, optimize=["einsum_path", (0, 1, 2)])
    )

    pbar = ms_tqdm(
        total=num_steps,
        bar_format=_progress_fmt,
        desc="Computing backward recursion.",
        disable=False if verbose is True else True,
    )

    # Initialize the backward recursion.
    if out is None:
        out = np.zeros((num_steps, len(Y)))
    out[num_steps - 1, :] = indicator_fn(Y, target_tube[num_steps - 1])
    pbar.update()

    # Backward recursion.
    for t in range(num_steps - 2, -1, -1):

        wX = np.einsum(
            "i,ijk->jk",
            value_functions[t + 1, :],
            betaXY,
            optimize=["einsum_path", (0, 1)],
        )

        VX = np.max(wX, axis=1)
        out[t, :] = step_fn(Y, VX, constraint_tube[t], target_tube[t])

        pbar.update()

    pbar.close()

    return out


def _compute_backward_recursion_batch(
    Y,
    W,
    CXY,
    CUA,
    num_steps=None,
    constraint_tube=None,
    target_tube=None,
    value_functions=None,
    out=None,
    step_fn=None,
    batch_size=None,
    verbose=False,
):
    pbar = ms_tqdm(
        total=num_steps,
        bar_format=_progress_fmt,
        desc="Computing backward recursion.",
        disable=False if verbose is True else True,
    )

    # Initialize the backward recursion.
    if out is None:
        out = np.zeros((num_steps, len(Y)))
    out[num_steps - 1, :] = indicator_fn(Y, target_tube[num_steps - 1])
    pbar.update()

    # Backward recursion.
    for t in range(num_steps - 2, -1, -1):

        for batch in generate_batches(len(Y), batch_size=batch_size):

            # Compute beta.
            betaXY = normalize(
                np.einsum(
                    "ii,ij,ik->ijk",
                    W,
                    CXY[:, batch],
                    CUA,
                    optimize=["einsum_path", (0, 1, 2)],
                )
            )

            wX = np.einsum(
                "i,ijk->jk",
                value_functions[t + 1, :],
                betaXY,
                optimize=["einsum_path", (0, 1)],
            )

            VX = np.max(wX, axis=1)
            out[t, batch] = step_fn(Y[batch], VX, constraint_tube[t], target_tube[t])

        pbar.update()

    pbar.close()

    return out


def kernel_sr_max(
    S: np.ndarray,
    A: np.ndarray,
    T: np.ndarray,
    num_steps: int = None,
    constraint_tube: list = None,
    target_tube: list = None,
    problem: str = "THT",
    regularization_param: float = None,
    kernel_fn=None,
    batch_size: int = None,
    verbose: bool = False,
):
    """Stochastic reachability using kernel distribution embeddings.

    Computes an approximation of the maximal safety probabilities of the stochastic
    reachability problem using kernel methods.

    Args:
        S: Sample of (x, u, y) tuples taken iid from the system evolution. The
            sample should be in the form of a list of tuples.
        A: Collection of admissible control action to choose from. Should be in the
            form of a 2D-array, where each row indicates a point.
        T: Evaluation points to evaluate the safety probabilities at. Should be in the
            form of a 2D-array, where each row indicates a point.
        num_steps: Number of time steps to compute the approximation.
        constraint_tube: List of spaces or constraint functions. Must be the same
            length as `num_steps`.
        target_tube: List of spaces or target functions. Must be the same length as
            `num_steps`.
        problem: One of `{"THT", "FHT"}`. `"THT"` specifies the terminal-hitting time
            problem and `"FHT"` specifies the first-hitting time problem.
        kernel_fn: Kernel function used by the approximation.
        regularization_param: Regularization parameter used in the solution to the
            regularized least-squares problem.
        batch_size: The batch size for more memory-efficient computations. Omit this
            parameter or set to `None` to compute without batch processing.
        verbose: Boolean flag to indicate verbose output.

    Returns:
        An array of safety probabilities of shape {len(T), num_steps}, where each row
        indicates the safety probabilities of the evaluation points at a different time
        step.

    """

    alg = KernelMaximalSR(
        num_steps=num_steps,
        constraint_tube=constraint_tube,
        target_tube=target_tube,
        problem=problem,
        regularization_param=regularization_param,
        kernel_fn=kernel_fn,
        batch_size=batch_size,
        verbose=verbose,
    )
    alg.fit(S, A)
    return alg.predict(T)


class KernelMaximalSR(AlgorithmInterface):
    """Stochastic reachability using kernel distribution embeddings.

    Computes an approximation of the maximal safety probabilities of the stochastic
    reachability problem using kernel methods.

    Args:
        num_steps: Number of time steps to compute the approximation.
        constraint_tube: List of spaces or constraint functions. Must be the same
            length as `num_steps`.
        target_tube: List of spaces or target functions. Must be the same length as
            `num_steps`.
        problem: One of `{"THT", "FHT"}`. `"THT"` specifies the terminal-hitting time
            problem and `"FHT"` specifies the first-hitting time problem.
        kernel_fn: Kernel function used by the approximation.
        regularization_param: Regularization parameter used in the solution to the
            regularized least-squares problem.
        batch_size: The batch size for more memory-efficient computations. Omit this
            parameter or set to `None` to compute without batch processing.
        verbose: Boolean flag to indicate verbose output.

    """

    def __init__(
        self,
        num_steps: int = None,
        constraint_tube: list = None,
        target_tube: list = None,
        problem: str = "THT",
        regularization_param: float = None,
        kernel_fn=None,
        batch_size: int = None,
        verbose: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.num_steps = num_steps
        self.constraint_tube = constraint_tube
        self.target_tube = target_tube

        self.problem = problem

        self.kernel_fn = kernel_fn
        self.regularization_param = regularization_param

        self.batch_size = batch_size

        self.verbose = verbose

    def _validate_params(self, S, A):

        if self.kernel_fn is None:
            self.kernel_fn = partial(rbf_kernel, sigma=0.1)

        if self.regularization_param is None:
            self.regularization_param = 1

        if self.num_steps is None:
            raise ValueError("Must supply a num_steps.")

        if self.constraint_tube is None:
            raise ValueError("Must supply constraint tube.")

        if self.target_tube is None:
            raise ValueError("Must supply target tube.")

        if self.problem not in ("FHT", "THT"):
            raise ValueError(f"problem is not in {'THT', 'FHT'}")

        if self.problem == "THT":
            self.step_fn = _tht_step
        elif self.problem == "FHT":
            self.step_fn = _fht_step

        if self.batch_size is not None:
            self._compute_backward_recursion_caller = partial(
                _compute_backward_recursion_batch, batch_size=self.batch_size
            )

        else:
            self._compute_backward_recursion_caller = _compute_backward_recursion

    def _validate_data(self, S=None):

        if S is None:
            raise ValueError("Must supply a sample.")

    def fit(self, S: np.ndarray, A: np.ndarray):
        """Run the algorithm.

        Args:
            S: Sample of (x, u, y) tuples taken iid from the system evolution. The
                sample should be in the form of a list of tuples.
            A: Collection of admissible control action to choose from. Should be in
                the form of a 2D-array, where each row indicates a point.

        Returns:
            self : Instance of KernelMaximalSR class.

        """

        self._validate_params(S, A)
        self._validate_data(S)
        self._validate_data(A)

        X, U, Y = transpose_sample(S)
        X = np.array(X)
        U = np.array(U)
        Y = np.array(Y)

        self.X = X

        A = np.array(A)

        gym_socks.logger.debug("Computing matrix inverse.")
        self.W = regularized_inverse(
            X,
            U=U,
            kernel_fn=self.kernel_fn,
            regularization_param=self.regularization_param,
        )

        gym_socks.logger.debug("Computing covariance matrices.")
        CXY = self.kernel_fn(X, Y)
        self.CUA = self.kernel_fn(U, A)

        gym_socks.logger.debug("Computing value functions.")
        value_functions = np.empty((self.num_steps, len(Y)))
        self.value_functions = self._compute_backward_recursion_caller(
            Y,
            self.W,
            CXY,
            self.CUA,
            num_steps=self.num_steps,
            constraint_tube=self.constraint_tube,
            target_tube=self.target_tube,
            value_functions=value_functions,
            out=value_functions,
            step_fn=self.step_fn,
            verbose=True,
        )

    def predict(self, T):
        """Predict.

        Args:
            T: Evaluation points to evaluate the safety probabilities at. Should be in
                the form of a 2D-array, where each row indicates a point.

        Returns:
            An array of safety probabilities of shape {len(T), num_steps}, where each
            row indicates the safety probabilities of the evaluation points at a
            different time step.

        """

        self._validate_data(T)
        T = np.array(T)

        gym_socks.logger.debug("Computing covariance matrix.")
        CXT = self.kernel_fn(self.X, T)

        gym_socks.logger.debug("Computing safety probabilities.")
        safety_probabilities = self._compute_backward_recursion_caller(
            T,
            self.W,
            CXT,
            self.CUA,
            num_steps=self.num_steps,
            constraint_tube=self.constraint_tube,
            target_tube=self.target_tube,
            value_functions=self.value_functions,
            step_fn=self.step_fn,
            verbose=True,
        )

        return safety_probabilities



"""ND point mass system."""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np


class NDPointMassEnv(DynamicalSystem):
    """ND point mass system.

    A point mass is a very simple system in which the inputs apply directly to the state
    variables. Thus, it is essentially a representation of a particle using Newton's
    equations `F = mA`.

    Args:
        dim: Dimensionality of the point mass system.
        mass: Mass of the particle.

    """

    def __init__(self, dim, mass=1, *args, **kwargs):
        """Initialize the system."""
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        self.mass = mass

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        return 1e-2 * np.array(w)

    def dynamics(self, time, state, action, disturbance):
        return (1 / self.mass) * np.array([*action], dtype=np.float32) + disturbance



"""Forward reachability example.

This file demonstrates the forward reachability classifier on a set of dummy data. Note
that the data is not taken from a dynamical system, but can easily be adapted to data
taken from system observations via a simple substitution. The reason for the dummy data
is to showcase the technique on a non-convex forward reachable set.

Example:
    To run the example, use the following command:

        $ python -m examples.forward_reach.forward_reach

.. [1] `Learning Approximate Forward Reachable Sets Using Separating Kernels, 2021
        Adam J. Thorpe, Kendric R. Ortiz, Meeko M. K. Oishi
        Learning for Dynamics and Control,
        <https://arxiv.org/abs/2011.09678>`_

"""

import gym
import gym_socks

import logging

import numpy as np

from functools import partial
from sacred import Experiment

from gym_socks.algorithms.reach.separating_kernel import SeparatingKernelClassifier

from sklearn.metrics.pairwise import euclidean_distances

from gym_socks.envs.sample import sample

from examples._computation_timer import ComputationTimer
from examples.ingredients.forward_reach_ingredient import forward_reach_ingredient
from examples.ingredients.forward_reach_ingredient import generate_test_points

from examples.ingredients.plotting_ingredient import plotting_ingredient
from examples.ingredients.plotting_ingredient import update_rc_params


@forward_reach_ingredient.config
def forward_reach_config():

    test_points = {
        "lower_bound": -1,
        "upper_bound": 1,
        "grid_resolution": 100,
    }


ex = Experiment(
    ingredients=[
        forward_reach_ingredient,
        plotting_ingredient,
    ]
)


@ex.config
def config():
    """Experiment configuration variables.

    SOCKS uses sacred to run experiments in order to ensure repeatability. Configuration
    variables are parameters that are passed to the experiment, such as the random seed,
    and can be specified at the command-line.

    Example:
        To run the experiment normally, use:

            $ python -m <experiment>

        The full configuration can be viewed using:

            $ python -m <experiment> print_config

        To specify configuration variables, use `with variable=value`, e.g.

            $ python -m <experiment> with seed=123 system.time_horizon=5

    .. _sacred:
        https://sacred.readthedocs.io/en/stable/index.html

    """

    sigma = 0.1
    sample_size = 500

    regularization_param = 1 / sample_size

    results_filename = "results/data.npy"
    no_plot = False


@ex.main
def main(
    _log,
    seed,
    sigma,
    regularization_param,
    sample_size,
    results_filename,
    no_plot,
):
    """Main experiment."""

    np.random.seed(seed)

    @gym_socks.envs.sample.sample_generator
    def donut_sampler() -> tuple:
        """Sample generator.

        Sample generator that generates points in a donut-shaped ring around the origin.
        An example of a non-convex region.

        Yields:
            sample : A sample taken iid from the region.

        """

        r = np.random.uniform(low=0.5, high=0.75, size=(1,))
        phi = np.random.uniform(low=0, high=2 * np.pi, size=(1,))
        point = np.array([r * np.cos(phi), r * np.sin(phi)])

        yield tuple(np.ravel(point))

    # Sample the distribution.
    S = sample(sampler=donut_sampler, sample_size=sample_size)

    # Generate the test points.
    T = generate_test_points((2,))

    with ComputationTimer():

        kernel_fn = partial(
            gym_socks.kernel.metrics.abel_kernel,
            sigma=sigma,
            distance_fn=euclidean_distances,
        )

        # Construct the algorithm.
        alg = SeparatingKernelClassifier(
            kernel_fn=kernel_fn,
            regularization_param=regularization_param,
        )

        # Train and classify the test points.
        alg.fit(S)
        labels = alg.predict(T)

    if not np.any(labels):
        _log.warning("No test points classified within reach set.")

    with open(results_filename, "wb") as f:
        np.save(f, S)
        np.save(f, T)
        np.save(f, labels)

    if not no_plot:
        plot_results()


@plotting_ingredient.config_hook
def plot_config(config, command_name, logger):
    if command_name in {"main", "plot_results"}:
        return {
            "axes": {
                "xlabel": r"$x_1$",
                "ylabel": r"$x_2$",
                "xlim": (-1.1, 1.1),
                "ylim": (-1.1, 1.1),
            },
        }


@ex.command(unobserved=True)
def plot_results(results_filename, plot_cfg):
    """Plot the results of the experiement."""

    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    logging.getLogger("PIL").setLevel(logging.WARNING)

    import matplotlib

    matplotlib.use("Agg")
    update_rc_params(matplotlib, plot_cfg["rc_params"])

    import matplotlib.pyplot as plt

    with open(results_filename, "rb") as f:
        S = np.array(np.load(f))
        T = np.array(np.load(f))
        labels = np.load(f)

    fig = plt.figure()
    ax = plt.axes(**plot_cfg["axes"])

    points_in = T[labels == True]

    plt.scatter(
        points_in[:, 0],
        points_in[:, 1],
        color="C0",
        marker=",",
        s=(300.0 / fig.dpi) ** 2,
    )

    plt.scatter(
        S[:, 0],
        S[:, 1],
        color="r",
        marker=".",
        s=1,
    )

    # Plot support region.
    plt.gca().add_patch(plt.Circle((0, 0), 0.5, fc="none", ec="blue"))
    plt.gca().add_patch(plt.Circle((0, 0), 0.75, fc="none", ec="blue"))

    plt.savefig(plot_cfg["plot_filename"])


if __name__ == "__main__":
    ex.run_commandline()



"""Common utility functions for ingredients."""

import gym

import numpy as np


def assert_config_has_key(_config, key):
    # if key not in _config:
    #     raise KeyError(f"Configuration {_config} missing entry: '{key}'.")
    msg = f"Configuration {_config} missing entry: '{key}'."
    assert key in _config, msg


def parse_array(value: float or list, shape: tuple, dtype: type) -> list:
    """Utility function for parsing configuration variables.

    Parses values passed as either a scalar or list into an array of type `dtype`.

    Args:
        value: The value input to the config.
        shape: The shape of the resulting array.
        dtype: The type to cast to.

    Returns:
        An array of a particular length.

    """

    if value is str:
        if value == "inf":
            value = np.inf
        elif value == "-inf":
            value = -np.inf

    elif value is list:
        for i, item in enumerate(value):
            if item == "inf":
                value[i] = np.inf
            elif item == "-inf":
                value[i] = -np.inf

    result = np.asarray(value)

    if np.isscalar(value):
        result = np.full(shape, value, dtype=dtype)

    result = np.asarray(result, dtype=dtype)

    return result


def box_factory(lower_bound, upper_bound, shape: tuple, dtype: type) -> gym.spaces.Box:
    """Box space factory.

    Creates a `gym.spaces.Box` to be used by sampling functions.

    """

    lower_bound = parse_array(lower_bound, shape=shape, dtype=dtype)
    upper_bound = parse_array(upper_bound, shape=shape, dtype=dtype)

    _space = gym.spaces.Box(
        low=lower_bound,
        high=upper_bound,
        shape=shape,
        dtype=dtype,
    )

    return _space


def grid_ranges(
    space: gym.Space,
    grid_resolution: list or int,
) -> list:
    """Compute grid ranges."""

    lower_bound = space.low
    upper_bound = space.high

    grid_resolution = parse_array(grid_resolution, shape=space.shape, dtype=int)

    xi = []
    for i in range(space.shape[0]):
        points = np.linspace(lower_bound[i], upper_bound[i], grid_resolution[i])
        xi.append(points.tolist())

    return xi


def grid_sample_size(
    space: gym.Space,
    grid_resolution: list or int,
) -> int:
    """Compute grid sample size."""
    grid_resolution = parse_array(grid_resolution, shape=space.shape, dtype=int)
    return int(np.product(grid_resolution))



import unittest
from unittest import mock
from unittest.mock import patch

import logging

import gym

import gym_socks

import numpy as np

from examples.ingredients.system_ingredient import system_ingredient
from examples.ingredients.system_ingredient import make_system
from examples.ingredients.system_ingredient import set_system_seed


class TestSystemIngredient(unittest.TestCase):
    def test_system_config(cls):
        """Test the ingredient config."""
        ingredient = system_ingredient
        cls.assertEqual(len(ingredient.configurations), 1)


class TestMakeSystem(unittest.TestCase):
    def test_make_system(cls):
        """Test that make_system makes a system based on system_id string."""

        dummy_logger = logging.getLogger(__name__)
        env = make_system(
            system_id="2DIntegratorEnv-v0",
            time_horizon=None,
            sampling_time=None,
            _config=dict(),
            _log=dummy_logger,
        )

        cls.assertIsInstance(env, gym_socks.envs.integrator.NDIntegratorEnv)



"""Sample ingredient.

The sample ingredient is used when an experiment requires a sample from a system to be
generated. It provides configuration variables used to set up the sample scheme, the
shape of the sample space, as well as the sample size. It also provides functions used
to generate the samples used in the experiments.

The main configuration specifications for the sample are given by:

* The need to specify different types of sampling schemes.
* The need to specify the bounds of the sampling region.
* The need to specify the "density" of the samples or, alternatively, the sample size.

The sampling ingredient should then provide functions which:

* Generate the appropriate sampling spaces.
* Provide an interface to the sampling functions that provide a single point of entry.

Notes:
    The main problem is that sacred does not allow for "substituting" different
    configurations depending on other configuration choices. For instance, if a user
    specifies a certain sampling scheme, there is no explicit method to modify the other
    variables required, since the ingredients are not dynamic.

    Ideally, the user would be able to specify a certain sampling scheme, and an
    ingredient could be loaded and added to the experiment "dynamically", injecting new
    configuration variables that need to be specified for the particular sampling
    scheme and providing a consistent "interface" that can be used regardless of the
    chosen sampling scheme.

    However, sacred *does* allow for configuration "hooks" which means we can
    dynamically add configuration variables at runtime. Additionally, dictionary
    configuration variables in sacred are updated to include new values rather than
    overwriting the entire dictionary if the same configuration variable is specified
    multiple times. This means we can simulate the dynamic ingredients using this
    procedure as a workaround.

"""


from random import Random
import gym
import gym_socks

import numpy as np

from functools import wraps

from sacred import Ingredient

from gym_socks.envs.dynamical_system import DynamicalSystem
from gym_socks.envs.policy import BasePolicy
from gym_socks.envs.policy import ConstantPolicy
from gym_socks.envs.policy import RandomizedPolicy
from gym_socks.envs.policy import ZeroPolicy

from gym_socks.envs.sample import sample, sequential_action_sampler
from gym_socks.envs.sample import sample_generator
from gym_socks.envs.sample import step_sampler
from gym_socks.envs.sample import uniform_grid
from gym_socks.envs.sample import uniform_grid_step_sampler

from examples.ingredients.common import grid_sample_size, parse_array
from examples.ingredients.common import box_factory
from examples.ingredients.common import grid_ranges

sample_ingredient = Ingredient("sample")


@sample_ingredient.config
def _config():
    """Sample configuration variables.

    The default configuration sets the sample types for both the `sample_space` and
    `action_space` samplers.

    Config:
        action_space: The configuration settings for the action space. Used for
            generating admissible control actions.
        sample_space: The configuration settings for the sample space. Used for
            generating samples from a system.
        sample_policy: The policy used when generating samples from a system.

    """

    sample_space = {"sample_scheme": "uniform"}
    sample_policy = {"sample_scheme": "random"}

    action_space = {"sample_scheme": "uniform"}


@sample_ingredient.config_hook
def _setup_random_sample_space_config_hook(config, command_name, logger):
    """Random sample configuration.

    If `random` is specified for the `sample_space` or `action_space` configuration
    variables, a `sample_size` is also required.

    """

    sample = config["sample"]
    update = dict()

    _defaults = {
        "sample_size": 1000,
    }

    if sample["sample_space"]["sample_scheme"] == "random":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["sample_space"] = {**_defaults, **sample["sample_space"]}

    if sample["action_space"]["sample_scheme"] == "random":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["action_space"] = {**_defaults, **sample["action_space"]}

    return update


@sample_ingredient.config_hook
def _setup_uniform_sample_space_config_hook(config, command_name, logger):
    """Uniform sample configuration.

    If `uniform` is specified for the `sample_space` or `action_space` configuration
    variables, upper and lower bounds are required.

    Upper and lower bounds can be specified as either a scalar value, or as an array of
    values that has the same shape as the space.

    """

    sample = config["sample"]
    update = dict()

    _defaults = {
        "lower_bound": -1,
        "upper_bound": 1,
        "sample_size": 1000,
    }

    if sample["sample_space"]["sample_scheme"] == "uniform":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["sample_space"] = {**_defaults, **sample["sample_space"]}

    if sample["action_space"]["sample_scheme"] == "uniform":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["action_space"] = {**_defaults, **sample["action_space"]}

    return update


@sample_ingredient.config_hook
def _setup_grid_sample_space_config_hook(config, command_name, logger):
    """Grid sample configuration.

    If `grid` is specified for the `sample_space` or `action_space` configuration
    variables, upper and lower bounds are required, along with a grid "resolution".

    Upper and lower bounds can be specified as either a scalar value, or as an array of
    values that has the same shape as the space.

    The grid resolution can be specified as a scalar value or as an array of values that
    has the same shape as the space.

    """

    sample = config["sample"]
    update = dict()

    _defaults = {
        "lower_bound": -1,
        "upper_bound": 1,
        "grid_resolution": 10,
    }

    if sample["sample_space"]["sample_scheme"] == "grid":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["sample_space"] = {**_defaults, **sample["sample_space"]}

    if sample["sample_policy"]["sample_scheme"] == "grid":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["sample_policy"] = {**_defaults, **sample["sample_policy"]}

    if sample["action_space"]["sample_scheme"] == "grid":
        # Merge dictionaries, being careful not to overwrite existing entries.
        update["action_space"] = {**_defaults, **sample["action_space"]}

    return update


def _sample_space_factory(shape: tuple, space_config: dict):
    """Sample space factory.

    Creates a sample space based on configuration variables.

    """

    sample_scheme = space_config["sample_scheme"]

    if sample_scheme == "random":
        lower_bound = -np.inf
        upper_bound = np.inf
    else:
        lower_bound = space_config["lower_bound"]
        upper_bound = space_config["upper_bound"]

    _space = box_factory(lower_bound, upper_bound, shape, dtype=np.float32)
    return _space


def _get_sample_size(space: gym.spaces.Box, space_config: dict):
    """Gets the sample size from config variables."""

    if space_config["sample_scheme"] == "grid":
        return grid_sample_size(
            space=space, grid_resolution=space_config["grid_resolution"]
        )

    return space_config["sample_size"]


@sample_ingredient.capture
def _default_sampler(seed: int, env: DynamicalSystem, sample_space):
    """Default sampler.

    By default, the sample generator yields a random sample from the stste space. Since
    the space is a `gym.spaces.Box`, it yields a sample using different underlying
    distributions depending on the bounds of the space. For bounded spaces, the
    distribution is uniform.

    Args:
        seed: The random seed.
        env: The dynamical system model.
        sample_space: The sample space configuration variable.

    Returns:
        A sampler used by the `_sample_ingredient_sampler`.

    """

    _space = _sample_space_factory(env.state_space.shape, sample_space)
    _space.seed(seed=seed)

    @sample_generator
    def _sample_generator(*args, **kwargs):
        yield _space.sample()

    return _sample_generator


@sample_ingredient.capture
def _grid_sampler(seed: int, env: DynamicalSystem, sample_space):
    """Grid sampler.

    Returns points on a grid.

    Args:
        seed: Unused.
        env: The dynamical system model.
        sample_space: The sample space configuration variable.

    Returns:
        A sampler used by the `_sample_ingredient_sampler`.

    """

    _space = _sample_space_factory(env.state_space.shape, sample_space)

    @sample_generator
    def _sample_generator(*args, **kwargs):
        xi = grid_ranges(_space, sample_space["grid_resolution"])
        xc = uniform_grid(xi)
        for item in xc:
            yield item

    return _sample_generator


@sample_ingredient.capture
def _random_action_sampler(seed: int, env: DynamicalSystem):
    """Random policy sampler.

    Args:
        seed: Unused.
        env: The dynamical system model.

    Returns:
        A sampler used by the `_sample_ingredient_sampler`.

    """

    _policy = RandomizedPolicy(env)

    @sample_generator
    def _sample_generator(*args, **kwargs):
        yield _policy(*args, **kwargs)

    return _sample_generator


@sample_ingredient.capture
def _zero_action_sampler(seed: int, env: DynamicalSystem):
    """Zero policy sampler.

    Args:
        seed: Unused.
        env: The dynamical system model.

    Returns:
        A sampler used by the `_sample_ingredient_sampler`.

    """
    _policy = ZeroPolicy(env)

    @sample_generator
    def _sample_generator(*args, **kwargs):
        yield _policy(*args, **kwargs)

    return _sample_generator


@sample_ingredient.capture
def _sequential_action_sampler(
    seed: int,
    env: DynamicalSystem,
    sample_space: dict,
    sample_policy: dict,
):
    """Uniform action sampler.

    Generates a sample using multiple actions at a uniform grid of points taken from
    within the range specified by `sample_policy`. Note that this is a simplification to
    make the result appear more uniform, but is not necessary for the correct operation
    of the algorithm. A random iid sample taken from the action space is usually
    sufficient.

    Args:
        seed: Unused.
        env: The dynamical system model.
        sample_space: The sample space configuration variable.
        sample_policy: The sample policy configuration variable.

    Returns:
        A sampler used by the `_sample_ingredient_sampler`.

    """
    _space = _sample_space_factory(env.action_space.shape, sample_policy)

    sample_size = _get_sample_size(env.state_space, sample_space)

    @sample_generator
    def _sample_generator(*args, **kwargs):
        xi = grid_ranges(_space, sample_policy["grid_resolution"])
        xc = uniform_grid(xi)
        for item in xc:
            for i in range(sample_size):
                yield item

    return _sample_generator


@sample_ingredient.capture
def _sample_ingredient_sampler(
    seed: int,
    env: DynamicalSystem,
    sample_space: dict,
    sample_policy: dict,
):
    """Custom sampler for the sample ingredient.

    The sample ingredient requires a sampler which is highly modular, allowing for
    several combinations of config variables. Thus, it uses two generators, one for the
    state space, and another for the action space. These generators are then used within
    a "standard" step-sampler, that generates a (state, action, next_state) tuple.

    Args:
        seed: Unused.
        env: The dynamical system model.
        sample_space: The sample space configuration variable.
        sample_policy: The sample policy configuration variable.

    Returns:
        A sampler that is used by the `generate_sample` function.

    """
    _sampler_map = {
        "random": _default_sampler,
        "uniform": _default_sampler,
        "grid": _grid_sampler,
    }

    state_sampler = _sampler_map[sample_space["sample_scheme"]]
    state_sample_generator = state_sampler(seed=seed, env=env)()

    _action_sampler_map = {
        "random": _random_action_sampler,
        "zero": _zero_action_sampler,
        "grid": _sequential_action_sampler,
    }

    action_sampler = _action_sampler_map[sample_policy["sample_scheme"]]
    action_sample_generator = action_sampler(seed=seed, env=env)()

    @sample_generator
    def _sample_generator(*args, **kwargs):

        state = next(state_sample_generator)
        action = next(action_sample_generator)

        env.state = state
        next_state, cost, done, _ = env.step(action)

        yield (state, action, next_state)

    return _sample_generator


@sample_ingredient.capture
def generate_sample(
    seed: int,
    env: DynamicalSystem,
    sample_space: dict,
    sample_policy: dict,
    action_space: dict,
    _log,
):
    """Generate a sample based on the ingredient config.

    Generates a sample based on the ingredient configuration variables. For instance, if
    the `sample_space` key `"sample_scheme"` is `"random"`, then the initial conditions
    of the sample are chosen randomly. A similar pattern follows for the `action_space`.
    The `sample_policy` determines the type of policy applied to the system during
    sampling.

    Args:
        seed: Unused.
        env: The dynamical system model.
        sample_space: The sample space configuration variable.
        sample_policy: The sample policy configuration variable.
        action_space: The action_space configuration variable.

    Returns:
        A sample of observations taken from the system evolution.

    """

    _log.debug("Generating sample.")

    _sampler = _sample_ingredient_sampler(seed=seed, env=env)

    _sample_space = _sample_space_factory(env.state_space.shape, sample_space)
    sample_size = _get_sample_size(_sample_space, sample_space)

    if sample_policy["sample_scheme"] == "grid":
        _action_space = _sample_space_factory(env.action_space.shape, action_space)
        action_size = _get_sample_size(_action_space, sample_policy)
        sample_size = sample_size * action_size

    _log.debug(f"Sample size: {sample_size}")

    S = sample(
        sampler=_sampler,
        sample_size=sample_size,
    )

    return S


@sample_ingredient.capture
def generate_admissible_actions(
    seed: int,
    env: DynamicalSystem,
    action_space: dict,
    _log,
):
    """Generate a collection of admissible control actions."""

    _log.debug("Generating admissible control actions.")

    sample_scheme = action_space["sample_scheme"]

    _action_space = _sample_space_factory(env.action_space.shape, action_space)
    _action_space.seed(seed=seed)

    if sample_scheme in {"random", "uniform"}:
        A = []
        for i in range(action_space["sample_size"]):
            A.append(_action_space.sample())

        return A

    ui = grid_ranges(_action_space, action_space["grid_resolution"])
    A = uniform_grid(ui)

    return A



from abc import abstractmethod

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np

from scipy.constants import gravitational_constant


class CWHBase(object):
    """CWH base class.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from `CWHBase`, and define a custom
    `compute_state_matrix` and `compute_input_matrix` function.

    This class holds the shared parameters for the CWH systems, which include:

    * orbital radius
    * gravitational constant
    * celestial body mass
    * chief mass

    And provides methods to compute:

    * graviational parameter (mu)
    * angular velocity (n)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.time_horizon = 600  # [s]
        self.sampling_time = 20  # [s]

        # system parameters
        self._orbital_radius = 850 + 6378.1  # [m]
        # self._gravitational_constant = 6.673e-11
        self._gravitational_constant = gravitational_constant
        self._celestial_body_mass = 5.9472e24  # [kg]
        self._chief_mass = 300  # [kg]

        self.compute_mu()
        self.compute_angular_velocity()

    @property
    def orbital_radius(self):
        return self._orbital_radius

    @orbital_radius.setter
    def orbital_radius(self, value):
        self._orbital_radius = value
        self.compute_angular_velocity()

    @property
    def gravitational_constant(self):
        return self._gravitational_constant

    @gravitational_constant.setter
    def gravitational_constant(self, value):
        self._gravitational_constant = value
        self.compute_mu()

    @property
    def celestial_body_mass(self):
        return self._celestial_body_mass

    @celestial_body_mass.setter
    def celestial_body_mass(self, value):
        self._celestial_body_mass = value
        self.compute_mu()

    @property
    def chief_mass(self):
        return self._chief_mass

    @chief_mass.setter
    def chief_mass(self, value):
        self._chief_mass = value

    def compute_mu(self):
        self.mu = self.gravitational_constant * self.celestial_body_mass / 1e9
        self.compute_angular_velocity()

    def compute_angular_velocity(self):
        self.angular_velocity = np.sqrt(self.mu / (self.orbital_radius ** 3))

        self.state_matrix = self.compute_state_matrix(sampling_time=self.sampling_time)
        self.input_matrix = self.compute_input_matrix(sampling_time=self.sampling_time)

    @abstractmethod
    def compute_state_matrix(self, sampling_time):
        raise NotImplementedError

    @abstractmethod
    def compute_input_matrix(self, sampling_time):
        raise NotImplementedError


class CWH4DEnv(CWHBase, DynamicalSystem):
    """4D Clohessy-Wiltshire-Hill (CWH) system.

    The 4D CWH system is a simplification of the 6D dynamics to operate within a plane.
    Essentially, it ignores the 'z' component of the dynamics.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-0.01, high=0.01, shape=(2,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        self.state_matrix = self.compute_state_matrix(sampling_time=self.sampling_time)
        self.input_matrix = self.compute_input_matrix(sampling_time=self.sampling_time)

    def compute_state_matrix(self, sampling_time):
        n = self.angular_velocity
        nt = n * sampling_time
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        return [
            [4 - 3 * cos_nt, 0, (1 / n) * sin_nt, (2 / n) * (1 - cos_nt)],
            [
                6 * (sin_nt - nt),
                1,
                -(2 / n) * (1 - cos_nt),
                (1 / n) * (4 * sin_nt - 3 * nt),
            ],
            [3 * n * sin_nt, 0, cos_nt, 2 * sin_nt],
            [-6 * n * (1 - cos_nt), 0, -2 * sin_nt, 4 * cos_nt - 3],
        ]

    def compute_input_matrix(self, sampling_time):
        n = self.angular_velocity
        Ts = sampling_time
        nt = n * Ts
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        sin_nt2 = np.sin((nt) / 2)

        M = self.chief_mass

        eAt = [
            [
                4 * Ts - (3 * sin_nt) / n,
                0,
                (2 * (sin_nt2 ** 2)) / (n ** 2),
                -(2 * (sin_nt - nt)) / (n ** 2),
            ],
            [
                (12 * (sin_nt2 ** 2)) / n - 3 * (Ts ** 2) * n,
                Ts,
                (2 * (sin_nt - nt)) / (n ** 2),
                (8 * (sin_nt2 ** 2)) / (n ** 2) - (3 * (Ts ** 2)) / 2,
            ],
            [3 - 3 * cos_nt, 0, sin_nt / n, (4 * (sin_nt2 ** 2)) / n],
            [
                6 * sin_nt - 6 * nt,
                0,
                -(4 * (sin_nt2 ** 2)) / n,
                (4 * sin_nt) / n - 3 * Ts,
            ],
        ]

        B = [
            [0, 0],
            [0, 0],
            [(1 / M), 0],
            [0, (1 / M)],
        ]

        return np.matmul(eAt, B)

    def step(self, action, time=0):
        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        disturbance = self.generate_disturbance(time, self.state, action)

        # use closed-form solution
        self.state = (
            np.matmul(self.state_matrix, self.state)
            + np.matmul(self.input_matrix, action)
            + disturbance
        )

        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        w = np.multiply([1e-4, 1e-4, 5e-8, 5e-8], w)
        return np.array(w)

    def dynamics(self, time, state, action, disturbance):
        """
        Dynamics for the system.

        NOTE: The CWH system has a closed-form solution for the equations of
        motion, meaning the dynamics function presented here is primarily for
        reference. The scipy.solve_ivp function does not return the correct
        result for the dynamical equations, and will quickly run into numerical
        issues where the states explode. See the 'step' function for details
        regarding how the next state is calculated.
        """
        x1, x2, x3, x4 = state
        u1, u2 = action
        w1, w2, w3, w4 = disturbance

        dx1 = x1 + w1
        dx2 = x2 + w2
        dx3 = (
            3 * (self.angular_velocity ** 2) * x1
            + 2 * self.angular_velocity * x4
            + (u1 / self.chief_mass)
            + w3
        )
        dx4 = -2 * self.angular_velocity * x3 + (u2 / self.chief_mass) + w4

        return np.array([dx1, dx2, dx3, dx4], dtype=np.float32)


class CWH6DEnv(CWHBase, DynamicalSystem):
    """6D Clohessy-Wiltshire-Hill (CWH) system."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-0.01, high=0.01, shape=(3,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        self.state_matrix = self.compute_state_matrix(sampling_time=self.sampling_time)
        self.input_matrix = self.compute_input_matrix(sampling_time=self.sampling_time)

    def compute_state_matrix(self, sampling_time):
        n = self.angular_velocity
        nt = n * sampling_time
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        return [
            [4 - 3 * cos_nt, 0, 0, (1 / n) * sin_nt, (2 / n) * (1 - cos_nt), 0],
            [
                6 * (sin_nt - nt),
                1,
                0,
                -(2 / n) * (1 - cos_nt),
                (1 / n) * (4 * sin_nt - 3 * nt),
                0,
            ],
            [0, 0, cos_nt, 0, 0, (1 / n) * sin_nt],
            [3 * n * sin_nt, 0, 0, cos_nt, 2 * sin_nt, 0],
            [-6 * n * (1 - cos_nt), 0, 0, -2 * sin_nt, 4 * cos_nt - 3, 0],
            [0, 0, -n * sin_nt, 0, 0, cos_nt],
        ]

    def compute_input_matrix(self, sampling_time):
        n = self.angular_velocity
        Ts = sampling_time
        nt = n * Ts
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        sin_nt2 = np.sin((nt) / 2)

        M = self.chief_mass

        eAt = [
            [
                4 * Ts - (3 * sin_nt) / n,
                0,
                0,
                (2 * (sin_nt2 ** 2)) / (n ** 2),
                -(2 * (sin_nt - nt)) / (n ** 2),
                0,
            ],
            [
                (12 * (sin_nt2 ** 2)) / n - 3 * (Ts ** 2) * n,
                Ts,
                0,
                (2 * (sin_nt - nt)) / (n ** 2),
                (8 * (sin_nt2 ** 2)) / (n ** 2) - (3 * (Ts ** 2)) / 2,
                0,
            ],
            [0, 0, sin_nt / n, 0, 0, (2 * (sin_nt2 ** 2)) / (n ** 2)],
            [
                3 - 3 * cos_nt,
                0,
                0,
                sin_nt / n,
                (4 * (sin_nt2 ** 2)) / n,
                0,
            ],
            [
                6 * sin_nt - 6 * nt,
                0,
                0,
                -(4 * (sin_nt2 ** 2)) / n,
                (4 * sin_nt) / n - 3 * Ts,
                0,
            ],
            [0, 0, cos_nt - 1, 0, 0, sin_nt / n],
        ]

        B = [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [(1 / M), 0, 0],
            [0, (1 / M), 0],
            [0, 0, (1 / M)],
        ]

        return np.matmul(eAt, B)

    def step(self, action, time=0):
        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        disturbance = self.generate_disturbance(time, self.state, action)

        # use closed-form solution
        self.state = (
            np.matmul(self.state_matrix, self.state)
            + np.matmul(self.input_matrix, action)
            + disturbance
        )

        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        w = np.multiply([1e-4, 1e-4, 1e-4, 5e-8, 5e-8, 5e-8], w)
        return np.array(w)

    def dynamics(self, time, state, action, disturbance):
        """
        Dynamics for the system.

        NOTE: The CWH system has a closed-form solution for the equations of
        motion, meaning the dynamics function presented here is primarily for
        reference. The scipy.solve_ivp function does not return the correct
        result for the dynamical equations, and will quickly run into numerical
        issues where the states explode. See the 'step' function for details
        regarding how the next state is calculated.
        """
        x1, x2, x3, x4, x5, x6 = state
        u1, u2, u3 = action
        w1, w2, w3, w4, w5, w6 = disturbance

        dx1 = x1 + w1
        dx2 = x2 + w2
        dx3 = x3 + w3
        dx4 = (
            3 * (self.angular_velocity ** 2) * x1
            + 2 * self.angular_velocity * x5
            + (u1 / self.chief_mass)
            + w4
        )
        dx5 = -2 * self.angular_velocity * x4 + (u2 / self.chief_mass) + w5
        dx6 = -(self.angular_velocity ** 2) * x3 + (u3 / self.chief_mass) + w6

        return np.array([dx1, dx2, dx3, dx4, dx5, dx6], dtype=np.float32)



import unittest
from unittest import mock
from unittest.mock import patch

import gym

from gym_socks.envs import NDIntegratorEnv

import numpy as np

from examples.ingredients.backward_reach_ingredient import backward_reach_ingredient
from examples.ingredients.backward_reach_ingredient import generate_tube
from examples.ingredients.backward_reach_ingredient import compute_test_point_ranges


class TestBackwardReachIngredient(unittest.TestCase):
    def test_backward_reach_config(cls):
        """Test the ingredient config."""
        ingredient = backward_reach_ingredient
        cls.assertEqual(len(ingredient.configurations), 1)


class TestGenerateTube(unittest.TestCase):
    def test_generate_tube_scalar_bounds(cls):
        """Test that generate_tube creates a list of spaces when given scalar bounds."""

        env = NDIntegratorEnv(2)
        num_time_steps = env.num_time_steps

        tube_cfg = {
            "lower_bound": -1,
            "upper_bound": 1,
        }

        tube = generate_tube(env=env, bounds=tube_cfg)

        cls.assertEqual(len(tube), num_time_steps)

        for item in tube:

            cls.assertIsInstance(item, gym.spaces.Box)
            cls.assertTrue(np.array_equal(item.low, [-1, -1]))
            cls.assertTrue(np.array_equal(item.high, [1, 1]))
            cls.assertEqual(item.shape, (2,))

    def test_generate_tube_list_bounds(cls):
        """Test that generate_tube creates a list of spaces when given list bounds."""

        env = NDIntegratorEnv(2)
        num_time_steps = env.num_time_steps

        lower_bound = (-np.arange(num_time_steps)).tolist()
        upper_bound = np.arange(num_time_steps).tolist()

        tube_cfg = {
            "lower_bound": lower_bound,
            "upper_bound": upper_bound,
        }

        tube = generate_tube(env=env, bounds=tube_cfg)

        cls.assertEqual(len(tube), num_time_steps)

        for i, item in enumerate(tube):

            cls.assertIsInstance(item, gym.spaces.Box)
            cls.assertTrue(
                np.array_equal(
                    item.low, np.full(item.shape, lower_bound[i], dtype=np.float32)
                )
            )
            cls.assertTrue(
                np.array_equal(
                    item.high, np.full(item.shape, upper_bound[i], dtype=np.float32)
                )
            )
            cls.assertEqual(item.shape, (2,))

    def test_generate_tube_matrix_bounds(cls):
        """Test that generate_tube creates a list of spaces when given matrix bounds."""

        env = NDIntegratorEnv(2)
        num_time_steps = env.num_time_steps

        lower_bound = np.full((num_time_steps, 2), -1, dtype=np.float32).tolist()
        upper_bound = np.full((num_time_steps, 2), 1, dtype=np.float32).tolist()

        tube_cfg = {
            "lower_bound": lower_bound,
            "upper_bound": upper_bound,
        }

        tube = generate_tube(env=env, bounds=tube_cfg)

        cls.assertEqual(len(tube), num_time_steps)

        for i, item in enumerate(tube):

            cls.assertIsInstance(item, gym.spaces.Box)
            cls.assertTrue(np.array_equal(item.low, [-1, -1]))
            cls.assertTrue(np.array_equal(item.high, [1, 1]))
            cls.assertEqual(item.shape, (2,))


class TestComputeTestPointRanges(unittest.TestCase):
    def test_compute_test_point_ranges(cls):
        """Test that compute_test_point_ranges gives correct ranges."""

        env = NDIntegratorEnv(2)

        test_points = {
            "lower_bound": -1,
            "upper_bound": 1,
            "grid_resolution": 25,
        }

        result = compute_test_point_ranges(env=env, test_points=test_points)

        groundTruth = [np.linspace(-1, 1, 25), np.linspace(-1, 1, 25)]

        cls.assertTrue(np.array_equal(result, groundTruth))



import unittest

import gym

import gym_socks.utils

import numpy as np


class TestNormalize(unittest.TestCase):
    def test_normalize(cls):
        """Test normalize function."""

        # Single point.
        points = [1.0, 2.0]
        groundTruth = [0.33333333, 0.66666667]
        normalize_result = gym_socks.utils.normalize(points)
        cls.assertTrue(np.allclose(normalize_result, groundTruth))

        # Multiple points.
        points = [[1.0, 2.0], [0.5, 1.0]]
        groundTruth = [[0.66666667, 0.66666667], [0.33333333, 0.33333333]]
        normalize_result = gym_socks.utils.normalize(points)
        cls.assertTrue(np.allclose(normalize_result, groundTruth))

        points = [[1.0, 0.5], [2.0, 1.0]]
        groundTruth = [[0.33333333, 0.33333333], [0.66666667, 0.66666667]]
        normalize_result = gym_socks.utils.normalize(points)
        cls.assertTrue(np.allclose(normalize_result, groundTruth))


class TestIndicatorFunction(unittest.TestCase):
    def test_indicator_function_on_box(cls):
        """Test indicator function on box."""

        interval = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        # One point inside.
        points = [[0.1, 0.1]]
        groundTruth = np.array([True], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, set=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

        # Multiple points inside.
        points = [[0.1, 0.1], [0.2, -0.2]]
        groundTruth = np.array([True, True], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, set=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

        # One point in, one point out.
        points = [[0.1, 0.1], [1.2, -0.2]]
        groundTruth = np.array([True, False], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, set=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

    def test_indicator_function_on_function(cls):
        """Test indicator function on function."""

        def interval(points):
            return np.array(np.all(np.abs(points) <= 1, axis=1), dtype=bool)

        # One point inside.
        points = [[0.1, 0.1]]
        groundTruth = np.array([True], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, set=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

        # Multiple points inside.
        points = [[0.1, 0.1], [0.2, -0.2]]
        groundTruth = np.array([True, True], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, set=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

        # One point in, one point out.
        points = [[0.1, 0.1], [1.2, -0.2]]
        groundTruth = np.array([True, False], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, set=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))


class TestGenerateBatches(unittest.TestCase):
    """Test generate_batches."""

    def test_generate_batches(cls):
        """Test generate batches."""
        idx = np.arange(10)

        batches = gym_socks.utils.generate_batches(num_elements=10, batch_size=2)
        for i, batch in enumerate(batches):
            cls.assertTrue(len(idx[batch]) == 2)
            cls.assertTrue(np.all(np.equal(idx[batch], idx[i * 2 : i * 2 + 2])))

        batches = gym_socks.utils.generate_batches(num_elements=10, batch_size=5)
        for i, batch in enumerate(batches):
            cls.assertTrue(len(idx[batch]) == 5)
            cls.assertTrue(np.all(np.equal(idx[batch], idx[i * 5 : i * 5 + 5])))

    def test_batch_larger_than_set(cls):
        """Ensure batches past the maximum size work properly."""

        idx = np.arange(10)

        batches = gym_socks.utils.generate_batches(num_elements=14, batch_size=7)
        batches.send(None)
        batch = batches.send(None)
        cls.assertTrue(len(idx[batch]) == 3)
        cls.assertTrue(np.all(np.equal(idx[batch], idx[-3:])))

        batches = gym_socks.utils.generate_batches(num_elements=10, batch_size=14)
        batch = batches.send(None)
        cls.assertTrue(len(idx[batch]) == 10)
        cls.assertTrue(np.all(np.equal(idx[batch], idx[:])))



"""TORA (translational oscillation with rotational actuation) system."""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np


class TORAEnv(DynamicalSystem):
    """TORA (translational oscillation with rotational actuation) system.

    The TORA system is a mass with an attached pendulum (rotational oscillator) attached via a spring to a surface. This system is useful for modeling a variant of the pendulum system or a cart-pole system. The input is to the pendulum.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
            ),
            state_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            *args,
            **kwargs
        )

        # system parameters
        self._damping_coefficient = 0.1

    @property
    def damping_coefficient(self):
        return self._damping_coefficient

    @damping_coefficient.setter
    def damping_coefficient(self, value):
        self._damping_coefficient = value

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        return 1e-2 * np.array(w)

    def dynamics(self, time, state, action, disturbance):
        x1, x2, x3, x4 = state
        w1, w2, w3, w4 = disturbance

        dx1 = x2 + w1
        dx2 = -x1 + self.damping_coefficient * np.sin(x3) + w2
        dx3 = x4 + w3
        dx4 = action + w4

        return np.array([dx1, dx2, dx3, *dx4], dtype=np.float32)



from typing import ForwardRef
import unittest
from unittest import mock
from unittest.mock import patch

import os

import gym

from gym_socks.envs import NDIntegratorEnv

import numpy as np

from examples.ingredients.plotting_ingredient import plotting_ingredient
from examples.ingredients.plotting_ingredient import update_rc_params


class TestForwardReachIngredient(unittest.TestCase):
    def test_plotting_config(cls):
        """Test the ingredient config."""
        ingredient = plotting_ingredient
        cls.assertEqual(len(ingredient.configurations), 1)

    def test_rc_params_update(cls):
        """Test that update_rc_params changes the global parameters for matplotlib."""

        rc_params_filename = os.path.abspath("examples/ingredients/matplotlibrc")
        rc_params = {"lines": {"linestyle": ":"}}

        import matplotlib

        cls.assertEqual(matplotlib.rcParams["lines.linestyle"], "-")
        cls.assertEqual(matplotlib.rcParams["lines.linewidth"], 1.5)

        update_rc_params(
            matplotlib,
            rc_params=rc_params,
            rc_params_filename=rc_params_filename,
        )

        cls.assertEqual(matplotlib.rcParams["lines.linestyle"], ":")
        cls.assertEqual(matplotlib.rcParams["lines.linewidth"], 0.75)



import unittest
from unittest.mock import patch

import gym

import gym_socks.envs

from gym_socks.envs.policy import BasePolicy, ZeroPolicy

import numpy as np


class TestBasePolicy(unittest.TestCase):
    @classmethod
    @patch("gym_socks.envs.policy.BasePolicy.__abstractmethods__", set())
    def setUpClass(cls):
        cls.policy = BasePolicy()

    def test_default_policy_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            cls.policy()


class TestConstantPolicy(unittest.TestCase):
    @classmethod
    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def setUpClass(cls):
        cls.system = gym_socks.envs.dynamical_system.DynamicalSystem(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            seed=1,
        )

    def test_constant_policy_returns_constants(cls):
        policy = gym_socks.envs.policy.ConstantPolicy(system=cls.system, constant=1)
        cls.assertEqual(policy(), [1])

        policy = gym_socks.envs.policy.ConstantPolicy(system=cls.system, constant=5.0)
        cls.assertEqual(policy(), [5.0])


class TestZeroPolicy(unittest.TestCase):
    @classmethod
    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def setUpClass(cls):
        cls.system = gym_socks.envs.dynamical_system.DynamicalSystem(
            observation_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            action_space=gym.spaces.Box(
                low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
            ),
            seed=1,
        )

    def test_zero_policy_returns_zero(cls):
        policy = gym_socks.envs.policy.ZeroPolicy(system=cls.system)
        cls.assertEqual(policy(), [0])



from abc import ABC, abstractmethod

import numpy as np


class BasePolicy(ABC):
    """Base policy class.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from BasePolicy.

    The `__call__` method is the main point of entry for the policy classes. All
    subclasses must implement a `__call__` method. This makes the class callable, so
    that policies can be evaluated as::

        u = policy(x)

    Note:
        Policies come in four main varieties:

        * Time-invariant open-loop policies.
        * Time-invariant closed-loop policies.
        * Time-varying open-loop policies.
        * Time-varying closed-loop policies.

        Thus, the arguments to the `__call__` method should allow for `time` and
        `state` to be specified (if needed), and should be optional kwargs, meaning
        they should have a `None` default value, like so:

            >>> def __call__(self, time=None, state=None):
            ...     ...

    """

    def __init__(self, *args, **kwargs):
        ...

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """Evaluate the policy.

        Returns:
            action: An action returned by the policy.

        """
        raise NotImplementedError


class RandomizedPolicy(BasePolicy):
    """Randomized policy.

    A policy which returns a random control action.

    Args:
        system: The system the policy is defined on. Needed to specify the shape of
            the inputs and outputs.

    """

    def __init__(self, system, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.system = system

    def __call__(self, *args, **kwargs):
        return self.system.action_space.sample()


class ConstantPolicy(BasePolicy):
    """Constant policy.

    A policy which returns a constant control action.

    Args:
        system: The system the policy is defined on. Needed to specify the shape of
            the inputs and outputs.
        constant: The constant value returned by the policy.

    """

    def __init__(self, system, constant=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.system = system
        self.constant = constant

    def __call__(self, *args, **kwargs):
        return np.full(
            self.system.action_space.shape,
            self.constant,
            dtype=self.system.action_space.dtype,
        )


class ZeroPolicy(ConstantPolicy):
    """Zero policy.

    A policy which returns a constant (zero) control action.

    Args:
        system : The system the policy is defined on. Needed to specify the shape of
            the inputs and outputs.

    """

    def __init__(self, system, *args, **kwargs):
        super().__init__(system, *args, **kwargs)



__all__ = [
    "maximally_safe",
    "monte_carlo",
    "random_fourier_features",
    "stochastic_reachability",
]

from gym_socks.algorithms.reach.kernel_sr_max import KernelMaximalSR
from gym_socks.algorithms.reach.kernel_sr_max import kernel_sr_max
from gym_socks.algorithms.reach.kernel_sr import KernelSR
from gym_socks.algorithms.reach.kernel_sr import kernel_sr
from gym_socks.algorithms.reach.separating_kernel import SeparatingKernelClassifier



"""Common functions for kernel control algorithms.

This file contains common functions used by the kernel optimal control algorithms, and
implements an LP solver to compute the probability vector gamma. This functionality is
accessed via the `compute_solution` function, which serves as a single entrypoint, and
the unconstrained version is chosen if the constraint matrix D is None.

References:
    .. [1] `Stochastic Optimal Control via
            Hilbert Space Embeddings of Distributions, 2021
            Adam J. Thorpe, Meeko M. K. Oishi
            IEEE Conference on Decision and Control,
            <https://arxiv.org/abs/2103.12759>`_

"""
import gym_socks
import numpy as np

from scipy.optimize import linprog


def _check_constraint_matrix(D):
    return np.any(np.negative(D))


def compute_solution(
    C: np.ndarray, D: np.ndarray = None, heuristic: bool = False
) -> np.ndarray:
    """Compute the solution to the LP.

    Computes a solution to the linear program, choosing either to delegate to the unconstrained or constrained solver depending on whether D is `None`.

    Args:
        C: Array holding values of the cost function evaluated at sample points.
        D: Array holding values of the constraint function evaluated at sample points.
        heuristic: Whether to compute the heuristic solution.

    Returns:
        gamma: Probability vector.
    """
    if D is None:
        return _compute_unconstrained_solution(C, heuristic)
    else:
        return _compute_constrained_solution(C, D, heuristic)


def _compute_unconstrained_solution(C: np.ndarray, heuristic=False) -> np.ndarray:
    """Compute the unconstrained solution to the LP.

    NOTE: A heuristic solution is available, but due to the speed of the LP solver for
    feasible problems, there is usually no need to invoke the heuristic solution, and
    the computation times are roughly equal. The main reason to use the heuristic
    solution is when scipy is unavailable.

    Note that in the unconstrained case, a closed-form solution is available by the
    Lagrangian dual. Thus, using the heuristic solution can be much faster in the
    unconstrained case.

    Args:
        C: Array holding values of the cost function evaluated at sample points.
        heuristic: Whether to compute the heuristic solution.

    Returns:
        gamma: Probability vector.

    """
    # C = (Cx @ K + Cu)

    if heuristic is False:

        obj = C.T
        A_eq = np.ones((1, len(C)))
        b_eq = 1
        # Bounds are automatically set so that decision variables are nonnegative.
        # bounds = [(0, None)] * len(C)

        gym_socks.logger.debug("Computing solution via scipy LP solver.")
        sol = linprog(
            obj,
            A_eq=A_eq,
            b_eq=b_eq,
        )

        gym_socks.logger.debug(f"Solver completed with status code: {sol.status}")
        # 0 : Optimization terminated successfully.
        # 1 : Iteration limit reached.
        # 2 : Problem appears to be infeasible.
        # 3 : Problem appears to be unbounded.
        # 4 : Numerical difficulties encountered.

        if sol.success is True:
            return sol.x
        else:
            gym_socks.logger.warn("No solution found via scipy.optimize.linprog.")
            gym_socks.logger.warn("Returning heuristic solution.")

    gym_socks.logger.debug("Computing heuristic solution.")
    heuristic_sol = np.zeros((len(C),))
    idx = np.argmin(C)
    heuristic_sol[idx] = 1
    return heuristic_sol


def _compute_constrained_solution(
    C: np.ndarray, D: np.ndarray, heuristic=False
) -> np.ndarray:
    """Compute the constrained solution to the LP.

    NOTE: A heuristic solution is available, but due to the speed of the LP solver for
    feasible problems, there is usually no need to invoke the heuristic solution, and
    the computation times are roughly equal. The main reason to use the heuristic
    solution is when scipy is unavailable.

    For the constrained problem, there is no closed-form solution via the Lagrangian
    dual. Thus, the heuristic solution computes the probability vector by masking the
    entries of D which are positive (solutions which do not satisfy the constraints),
    and then finding the minimum value of the masked cost vector.

    Args:
        C: Array holding values of the cost function evaluated at sample points.
        D: Array holding values of the constraint function evaluated at sample points.
        heuristic: Whether to compute the heuristic solution.

    Returns:
        gamma: Probability vector.

    """
    # C = (Cx @ K + Cu)
    # D = (Dx @ K + Du)
    _check_constraint_matrix(D)

    if heuristic is False:

        if len(D.shape) == 1:
            D = D.reshape(-1, 1)

        obj = C.T
        A_ub = D.T
        b_ub = 0
        A_eq = np.ones((1, len(C)))
        b_eq = 1
        # Bounds are automatically set so that decision variables are nonnegative.
        # bounds = [(0, None)] * len(C)

        gym_socks.logger.debug("Computing solution via scipy LP solver.")
        sol = linprog(
            obj,
            A_ub=A_ub,
            b_ub=b_ub,
            A_eq=A_eq,
            b_eq=b_eq,
        )

        gym_socks.logger.debug(f"Solver completed with status code: {sol.status}")
        # 0 : Optimization terminated successfully.
        # 1 : Iteration limit reached.
        # 2 : Problem appears to be infeasible.
        # 3 : Problem appears to be unbounded.
        # 4 : Numerical difficulties encountered.

        if sol.success is True:
            return sol.x
        else:
            gym_socks.logger.debug("No solution found via scipy.optimize.linprog.")
            gym_socks.logger.debug("Returning heuristic solution.")

    heuristic_sol = np.zeros((len(C),))
    satisfies_constraints = np.where(D <= 0)
    if len(satisfies_constraints[0]) == 0:
        gym_socks.logger.warn("No feasible solution found!")
        gym_socks.logger.debug("Returning minimal unconstrained solution.")
        idx = C.argmin()
    else:
        idx = satisfies_constraints[0][C[satisfies_constraints].argmin()]
    heuristic_sol[idx] = 1

    return heuristic_sol



import unittest
from unittest import mock
from unittest.mock import patch

import gym

import numpy as np

from examples.ingredients.tracking_ingredient import tracking_ingredient


class TestTrackingIngredient(unittest.TestCase):
    def test_tracking_config(cls):
        """Test the ingredient config."""
        ingredient = tracking_ingredient
        cls.assertEqual(len(ingredient.configurations), 1)
